package org.uclic.algebra;

import java.io.*;
import java.util.Date;
import java.util.Iterator;
import java.util.Vector;
import org.uclic.matrix.Matrix;
import org.uclic.matrix.WrongSize;
import org.uclic.algebra.*;
import org.uclic.algebra.event.*;

/**
 * Generates equivalence classes for products of an event algebra.
 *
 * @author Jeremy Gow
 * @version 0.4
 */
public class EquivClassGenerator {

    // Algebra
    private EventAlgebra  algebra;      // The event algebra
    private EventTerm[][] equivClasses; // Its equivalence classes
    private Matrix[]      matrices;     // Matrix for each class
    private int           classes;      // The number of classes

    // Search
    private int[]   uniqueEvents;         // Events used to build products
    private int     maxDepth;             // Maximum search depth
    private Vector  active, newActive; // Active class sets
    private boolean complete;             // Was the classification complete?
    private boolean maxed;                // Was the class limit reached?

    private StringBuffer report;

    private final static int MAX_DEPTH = 10;
    private final static int MAX_CLASS = 2000;
    
    public EquivClassGenerator(EventAlgebra a, int max) {
	algebra = a;
	if (max > MAX_DEPTH) {
	    maxDepth = MAX_DEPTH;
	} else {
	    maxDepth = max;
	}
    }

    public EventTerm[] getEquivClass(int i) {
	return equivClasses[i];
    }
    
    public Matrix getMatrix(int i) {
	return matrices[i];
    }

    public int classes() {
	return classes;
    }

    public boolean complete() {
	return complete;
    }

    public boolean classLimitReached() {
	return maxed;
    }


    /**
     * Generate equivalence classes for event algebra, up to given product length.
     */ 
    public void generate() {
	
	equivClasses = new EventTerm[MAX_CLASS][];
	matrices = new Matrix[MAX_CLASS];
	classes = 0;
	newActive = new Vector();
	complete = false;
	maxed = false;

	initialReport();

	addSpecialEvents();
	addEvents();
	addProducts();

	finalReport();
    }


    /**
     * Add an equivalence class with given representative and matrix.
     */
    private void addEquivClass(EventTerm rep, Matrix matrix) {
	if (classes < MAX_CLASS) {
	    report("Adding equiv class " + rep.toString());
	    equivClasses[classes] = new EventTerm[1];
	    equivClasses[classes][0] = rep;
	    matrices[classes] = matrix;
	    classes++;
	} else {
	    report("Equiv class array full.");
	    maxed = true;
	}
    }

    /**
     * Add an equivalence class with the given representative.
     */ 
    private void addEquivClass(EventTerm rep) {
	try {
	    addEquivClass(rep, rep.evaluate());
	} catch (EvaluationException me) {}
    }

    /**
     * Add an expression to its equivalence class, generating a new class if necessary.
     */
    private boolean addExpression(EventTerm expr, Matrix matrix) {
	
	boolean original = true;
	EventTerm[] newProds;
	EventTerm rep;

	// Check if in known class
	for (int i = 0; i < classes; i++) {
	    if (matrix.equals(matrices[i])) {
		report("Adding " + expr.toString() + " to equiv class " + equivClasses[i][0].toString());
		newProds = new EventTerm[equivClasses[i].length + 1];
		System.arraycopy(equivClasses[i], 0, newProds, 0, equivClasses[i].length);
		newProds[equivClasses[i].length] = expr;
		equivClasses[i] = newProds;
		original = false;
		break;
	    }
	}
	
	// Add new class if not already known
	if (original) {
	    newActive.add(new Integer(classes));
	    addEquivClass(expr, matrix);
	}

	return original;
    }

    /**
     * Add some predefined algebra events as equivalence classes.
     */
    private void addSpecialEvents() {
	addEquivClass(new Id(algebra));
	//addEquivClass(new Impossible());
	//addEquivClass(new Anything());
    }

    /**
     * Try and add the basic events to equivalence classes.
     */
    private void addEvents() {
	
       	UIEvent ui;
	int[] unique = new int[algebra.getModel().events().length];
	int i = 0;

	for (int e = 0; e < unique.length; e++) {
	    ui = new UIEvent(e, algebra);
	    try {
		if (addExpression(ui, ui.evaluate())) {
		    unique[i] = e;
		    i++;
		}	    
	    } catch (EvaluationException me) {}
	}

	if (i > 0) {
	    uniqueEvents = new int[i];
	    System.arraycopy(unique, 0, uniqueEvents, 0, i);
	}
    }	

    /**
     * Add event products to equivalence classes.
     */
    private void addProducts() {

	UIEvent ui;
	Iterator itr;
	int index;
	Matrix uiMatrix, prodMatrix;
	EventTerm expr, product;
	
	for (int c = 1; c < maxDepth; c++) {

	    report("\n\nLENGTH = " + c);

	    active = newActive;
	    itr = active.iterator();
	    newActive = new Vector();	    

	    while (itr.hasNext()) {
		index = ((Integer) itr.next()).intValue();
	
		for (int i = 0; i < uniqueEvents.length; i++) {

		    ui = new UIEvent(uniqueEvents[i], algebra);
		    EventTerm[] draft = new EventTerm[2];
		    draft[0] = equivClasses[index][0];
		    draft[1] = ui;
		    product = normalise(new EventProduct(draft, algebra));
		    report("\nPRODUCT: " + product.toString());

		    try {
			uiMatrix = ui.evaluate();
			prodMatrix = matrices[index].multiply(uiMatrix);
			
			if (!redundantSubterm(product)) {
			    addExpression(product, prodMatrix);
			} else {
			    report("Rejected product: " + product.toString());
			}
		    } catch (WrongSize ws) {
		    } catch (EvaluationException me) {}
		}
	    }

	    if (newActive.size() == 0) {
		complete = true;
		break;
	    }

	    Date date = new Date();
	    System.out.println("Product length " + c + " complete, " + date.toString());
	}
	 

    }

    private boolean redundantSubterm(EventTerm expr) {

	EventIterator itr1, itr2;
	boolean match;
	int len1 = expr.events();
	int len2, preBound;
	AtomicEvent e1, e2;

	for (int cls = 0; cls < classes; cls++) {

	    for (int i = 1; i < equivClasses[cls].length; i++) {

		report("   Looking for subterm " + equivClasses[cls][i].toString());

		len2 = equivClasses[cls][i].events();
		preBound = (len1 - len2) + 1;
		report("   Prefix bound = " + preBound);

		for (int pre = 0; pre < preBound; pre++) {

		    report("   Prefix = " + pre);

		    itr1 = expr.eventIterator();
		    itr2 = equivClasses[cls][i].eventIterator();

		    // Remove prefix from itr1
		    for (int j = 0; j < pre; j++) {
			itr1.next();
		    }

		    // Match terms
		    match = true;
		    while(itr1.hasNext() && itr2.hasNext()) {
			e1 = itr1.next();
			e2 = itr2.next();
			if (e1 != e2) {
			    match = false;
			    report("   Mismatch: " + e1 + " and " + e2);
			    break;
			}
		    }
		    if (match && !itr2.hasNext()) {
			return true;
		    }
		}
	    }
	}
	return false;
    }



    private EventTerm normalise(EventTerm expr) {

	Vector products = new Vector();
	EventTerm current, next, norm;
	EventIterator itr = expr.eventIterator();

	int power = 1;	
	current = itr.next();
		
	while (itr.hasNext()) {
	    
	    next = itr.next();
	    
	    if (next.equals(current)) {
		power++;

	    } else {
		if (power > 1) {
		    products.add(new Power(current, power, algebra));
		} else {
		    products.add(current);
		}
		power = 1;
		current = next;
	    }
	}
	if (power > 1) {
	    products.add(new Power(current, power, algebra));
	} else {
	    products.add(current);
	}

	if (products.size() > 1) {

	    norm = new EventProduct((EventTerm[]) products.toArray(), algebra);
	
	} else {

	    norm = (EventTerm) products.get(0);
	}

	return norm;
    }

    
    private void initialReport() {
	//report = new StringBuffer();
    }

    private void report(String str) {
	//report.append(str + "\n");
    } 

    private void finalReport() {
	/*
	try {
	    FileWriter writer = new FileWriter("/home/jeremy/maui/equiv_trace.txt");
            writer.write(report.toString());
            writer.close();
					
        } catch (IOException e) {
            e.printStackTrace();
        }
	*/
    }
}


