package org.uclic.algebra;

import java.io.*;
import java.util.*;
import org.jdom.*;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;
import org.uclic.matrix.*;
import org.uclic.fui.*;

/**
 * A mapping between event algebra elements and actual matrices and
 * vectors.  Event matrices and sign/superstate vectors are cached,
 * but state vectors are not.
 *
 * @version 0.4
 * @author Jeremy Gow
 *
 * Should lazily generate matrices.
 */
public class EventAlgebra {

    private SignFSM  model;
    private HashMap  eventMatrices;
    private HashMap  superVectors;
    private HashMap  signVectors;
    private Matrix   idMatrix;

    public EventAlgebra(SignFSM m) {
	super();
	model = m;
	eventMatrices = new HashMap();
	superVectors = new HashMap();
	signVectors = new HashMap();
    }

    public SignFSM getModel() {
	return model;
    }

    public Matrix identityMatrix() {

	if (idMatrix == null) {

	    int s = model.states().length;
	    idMatrix = new Matrix(s);
	    
	    for (int i = 0; i < s; i++) {
		for (int j = 0; j < s; j++) {
		    if (i == j) {
			idMatrix.setEntry(i, j, 1);
		    }
		}
	    }
	} 
	
	return idMatrix;
    }    


    public RowVector stateVector(int i) throws EvaluationException {
	RowVector vector = new RowVector(model.states().length);
	vector.setEntry(0, i, 1);
	return vector;
    }

    public Matrix eventMatrix(int i) throws EvaluationException {

	if (i > -1 && i < model.events().length) {

	    // Try cache
	    Integer bigI = new Integer(i);
	    Matrix em = (Matrix) eventMatrices.get(bigI);
	    
	    // Construct if not cached
	    if (em == null) {
		
		int s = model.states().length;
		em = new Matrix(s);

		// The jth from-state (row)
		for (int j = 0; j < s; j++) {
		    int[] trans = model.rangeIndices(j, i);
		    
		    if (trans != null) {
			// The trans[k]-th to-state (column)
			for (int k = 0; k < trans.length; k++) {
			    em.setEntry(j, trans[k], 1);
			}
		    }
		}
		eventMatrices.put(bigI, em);
	    }

	    return em;

	} else {
	    throw new EvaluationException("Invalid event index: " + i);
	}
    }

    public RowVector superVector(int i) throws EvaluationException {

	if (i > -1 && i < model.superstates().length) {

	    // Try cache
	    Integer bigI = new Integer(i);
	    RowVector sv = (RowVector) superVectors.get(bigI);
	    
	    // Construct if not cached
	    if (sv == null) {
		
		int s = model.states().length;
		sv = new RowVector(s);

		try {
		    // The jth state
		    for (int j = 0; j < s; j++) {
			if (model.isElement(j, i)) {
			    sv.setEntry(j, 1);
			}
		    }
		} catch (OutsideMatrix om) {}

		superVectors.put(bigI, sv);
	    }

	    return sv;

	} else {
	    throw new EvaluationException("Invalid superstate index: " + i);
	}
    }


    public RowVector signVector(int i) throws EvaluationException {

	if (i > -1 && i < model.signs().length) {

	    // Try cache
	    Integer bigI = new Integer(i);
	    RowVector sv = (RowVector) signVectors.get(bigI);
	    
	    // Construct if not cached
	    if (sv == null) {
		
		int s = model.states().length;
		sv = new RowVector(s);

		try {
		    // The jth state
		    for (int j = 0; j < s; j++) {
			if (model.isStateSign(j, i)) {
			    sv.setEntry(j, 1);
			}
		    }
		} catch (OutsideMatrix om) {}

		signVectors.put(bigI, sv);
	    }

	    return sv;

	} else {
	    throw new EvaluationException("Invalid sign index: " + i);
	}
    }

    public void redefined() {
	eventMatrices.clear();
	superVectors.clear();
	signVectors.clear();
	idMatrix = null;
    }

    public void deleteEvent(int i) {
	adjustCache(eventMatrices, i);
    }

    public void deleteState(int i) {
	// No state cache to adjust.
    }

    public void deleteSuperstate(int i) {
	adjustCache(superVectors, i);
    }

    public void deleteSign(int i) {
	adjustCache(signVectors, i);
    }


    private HashMap adjustCache(HashMap map, int i) {

	Iterator keys = map.keySet().iterator();
	Integer key;
	Object value;
	HashMap map2 = new HashMap();
	int j;

	while (keys.hasNext()) {
	    key = (Integer) keys.next();
	    value = map.get(key);
	    j = key.intValue();

	    if (i != j) {
		if (i < j) {
		    key = new Integer(j - 1);
		}
		map2.put(key, value);
	    }    
	}

	return map2;
    }

}
