package org.uclic.algebra;

import java.util.Hashtable;
import java.util.ListIterator;
import java.util.Iterator;
import java.util.Vector;
import org.jdom.Element;
import org.uclic.fui.*;
import org.uclic.algebra.event.*;
import org.uclic.algebra.property.*;
import org.uclic.algebra.state.*;
import org.uclic.util.*;

/**
 * The formal specification of a user interface.
 *
 * @author Jeremy Gow
 * @version 0.4
 *
 * To do: replace vectors with arrays.
 */
public class Specification {

    private Vector properties;
    private Vector values;
    private Vector percentages;
    private Vector locks;
    private Vector lockValues;
    private EventAlgebra algebra;

    public Specification(EventAlgebra a) {
	properties = new Vector();	
	values = new Vector();
	percentages = new Vector();
	locks = new Vector();
	lockValues = new Vector();
	algebra = a;
    }

    public Specification(EventAlgebra a, Specification s) {
	
	algebra = a;
       	int n = s.properties().size();
	
	properties = new Vector(n);
	for (int i = 0; i < n; i++) {
	    properties.add(s.property(i).copyTerm(a));
	}

	values = new Vector(n);
	for (int i = 0; i < n; i++) {
	    if (s.valueKnown(i)) {
		values.add(new Boolean(s.value(i)));
	    } else {
		values.add(null);
	    }
	}

	percentages = new Vector(n);
	for (int i = 0; i < n; i++) {
	    if (s.percentKnown(i)) {
		percentages.add(new Integer(s.percent(i)));
	    } else {
		percentages.add(null);
	    }
	}

	locks = new Vector(n);
	for (int i = 0; i < n; i++) {
	    locks.add(new Boolean(s.isLocked(i)));
	}
	
	lockValues = new Vector(n);
	for (int i = 0; i < n; i++) {
	    if (s.isLocked(i)) {
		lockValues.add(new Boolean(s.value(i)));
	    } else {
		lockValues.add(null);
	    }
	}	
    }

    public EventAlgebra algebra() {
	return algebra;
    }

    public void setAlgebra(EventAlgebra a) {
	algebra = a;
	Iterator itr = properties.iterator();
	while (itr.hasNext()) {
	    PropertyTerm prop = (PropertyTerm) itr.next();
	    prop.setAlgebra(a);
	}
    }

    // Size

    public int size() {
	return properties.size();
    }

    public boolean isEmpty() {
	return properties.isEmpty();
    }

    // Properties

    public void addProperty() {	
	addProperty(new EventEquality(algebra));
    }

    public void addProperty(PropertyTerm p) {
	int e = size();
	properties.add(p);
	values.add(null);
	percentages.add(null);
	locks.add(new Boolean(false));
	lockValues.add(null);
	evaluate(e);
    }

    public void addProperties(Vector p) {
	int e = size();
	int s = p.size();
	properties.addAll(p);
	for (int i = 0; i < s; i++) {
	    values.add(null);
	    percentages.add(null);
	    locks.add(new Boolean(false));
	    lockValues.add(null);
	    evaluate(e + i);
	}
    }

    public void addProperty(int i) {
	addProperty(new EventEquality(algebra), i);
    }

    public void addProperty(PropertyTerm p, int i) {
	properties.insertElementAt(p, i);
	values.insertElementAt(null, i);
	percentages.insertElementAt(null, i);
	locks.insertElementAt(new Boolean(false), i);
	lockValues.insertElementAt(null, i);
	evaluate(i);
    }

    public void removeProperty(int i) {
	properties.remove(i);
	values.remove(i);
	percentages.remove(i);
	locks.remove(i);
	lockValues.remove(i);
    }

    public void setProperty(PropertyTerm p, int i) {
	properties.setElementAt(p, i);
	evaluate(i);
    }

    public PropertyTerm property(int i) {
	return (PropertyTerm) properties.get(i);
    }

    public Vector properties() {
	return properties;
    }

    // Values

    public boolean valueKnown(int i) {
	return (values.get(i) != null);
    }

    public boolean value(int i) {
	Boolean result = (Boolean) values.get(i);
	return result.booleanValue();
    }

    public void allUnknown() {
	for (int i = 0; i < values.size(); i++) {
	    values.setElementAt(null, i);
	    percentages.setElementAt(null, i);
	}
    }

    public void evaluate() {	
	for (int i = 0; i < properties.size(); i++) {
	    evaluate(i);
	}
    }

    private void evaluate(int i) {

	Boolean value = null;
	Integer per = null;

	try {
	    boolean[] eval = property(i).evaluate();
	    int p = ArrayUtils.percent(eval);
	    value = new Boolean(p == 100);
	    per = new Integer(p);

	} catch (EvaluationException e) {}

	values.setElementAt(value, i);
	percentages.setElementAt(per, i);
    }


    // Percentages

    public boolean percentKnown(int i) {
	return (percentages.get(i) != null);
    }

    public int percent(int i) {
	Integer percent = (Integer) percentages.get(i);
	return percent.intValue();
    }

    public boolean partial(int i) {	
	return percentKnown(i) && (percent(i) != 0) && (percent(i) != 100);
    }

    // Locks

    public boolean isLocked(int i) {
	Boolean lock = (Boolean) locks.get(i);
	return lock.booleanValue();
    }

    public Boolean isLockedObject(int i) {
	Boolean lock = (Boolean) locks.get(i);
	return lock;
    }

    /*
     * Set lock on/off.  Can only be turned on when property's value is known.
     */
    public void setLock(int i, boolean val) {
	if (val) {
	    if (valueKnown(i)) {
		locks.setElementAt(new Boolean(true), i);
		lockValues.setElementAt(values.get(i), i);
	    }
	} else {
	    locks.setElementAt(new Boolean(false), i);
	    lockValues.setElementAt(null, i);
	}
    }

    /*
    private void evaluateLocked() {
	for (int i = 0; i < size(); i++) {
	    if (isLocked(i)) {
		evaluate(i);
	    }
	}
    }
    */

    public int[] checkLocks() {

	int[] broken = null;

	for (int i = 0; i < size(); i++) {
	    if (isLocked(i)) {
		//System.out.println("Property " + i + " is locked");
		boolean broke = false;
		boolean known = true;
		boolean current = true;
		
		try {
		    boolean[] eval = property(i).evaluate();
		    current = (ArrayUtils.percent(eval) == 100);
		} catch (EvaluationException e) {
		    known = false;
		}

		if (known) {

		    Boolean v = (Boolean) lockValues.get(i);

		    if (v.booleanValue() != current) {
			// Locked value is now changed.
			broke = true;
			//System.out.println("  - value now changed from " + v + " to " + current);
		    } else {
			//System.out.println("  - value remains " + current);
		    }
		} else {
		    // Locked value is now unknown.
		    broke = true;
		    //System.out.println("  - value now unknown");
		}

		if (broke) {
		    if (broken == null) {
			broken = new int[1];
			broken[0] = i;
		    } else {
			broken = ArrayUtils.addToArray(i, broken);
		    }
		}
	    } else {
		//System.out.println("Property " + i + " not locked");
	    }
	}

	return broken;
    }


    // Responseding to changes in model.

    public void deleteEvent(int i) {
	Iterator itr = properties.iterator();
	while (itr.hasNext()) {
	    PropertyTerm p = (PropertyTerm) itr.next();
	    p.deleteEvent(i);
	}
    }

    public void deleteState(int i) {
	Iterator itr = properties.iterator();
	while (itr.hasNext()) {
	    PropertyTerm p = (PropertyTerm) itr.next();
	    p.deleteState(i);
	}
    }

    public void deleteSuperstate(int i) {
	Iterator itr = properties.iterator();
	while (itr.hasNext()) {
	    PropertyTerm p = (PropertyTerm) itr.next();
	    p.deleteSuperstate(i);
	}
    }

    public void deleteSign(int i) {
	Iterator itr = properties.iterator();
	while (itr.hasNext()) {
	    PropertyTerm p = (PropertyTerm) itr.next();
	    p.deleteSign(i);
	}
    }


    // XML
    
    public Element toXML() {

	PropertyTerm p;
	Boolean b;
	Element elem, spec = new Element(FUI_XML.SPEC_ELEM);

	ListIterator pItr = properties.listIterator();
	ListIterator vItr = values.listIterator();

	while (pItr.hasNext()) {
	    p = (PropertyTerm) pItr.next();
	    b = (Boolean) vItr.next();
	    elem = new Element(FUI_XML.PROP_ELEM);
	    elem.addContent(p.toXML());
	    if (b != null) {
		elem.setAttribute("value", b.toString());
	    }	    
	    spec.addContent(elem);
	}

	return spec;
    }

}
