package org.uclic.algebra.event;

import java.util.List;
import java.util.NoSuchElementException;
import org.jdom.Element;
import org.uclic.fui.*;
import org.uclic.algebra.*;
import org.uclic.matrix.Matrix;
import org.uclic.matrix.WrongSize;

/**
 * An event raised to given power
 *
 * @version 0.4
 * @author Jeremy Gow
 */
public class Power extends EventTerm {
    
    public final static String PRE_STRING  = "(";
    public final static String IN_STRING   = ")^";
    public final static String NODE_STRING = "Power";

    public Power(EventTerm e, IntTerm n, EventAlgebra a) {
	super(a);
	add(e);
	add(n);
    }

    public Power(EventTerm e, int n, EventAlgebra a) {
	this(e, new IntTerm(n, a), a);
    }

    public Power(EventAlgebra a) {
	this(new UIEvent(a), 2, a);
    }

    public EventTerm getBase() {
	return (EventTerm) getSubterm(0);
    }

    public IntTerm getExponent() {
	return (IntTerm) getSubterm(1);
    }

    public Matrix evaluate() throws EvaluationException {

	Matrix em, m = null;
	int exp = getExponent().value();

	if (exp < 1) {
	    m = algebra.identityMatrix();
	} else {
	    m = getBase().evaluate();
	    em = m;
	    if (exp > 1) {
		for (int i = 1; i < exp; i++) {
		    try {
			m = m.multiply(em);
		    } catch (WrongSize e) {
			throw new EvaluationException("Event matrix cannot be raised to power.");
		    }
		}
	    }
	}
	return m;
    }

    public String termString() {
	return PRE_STRING + getSubterm(0).termString() + IN_STRING + getSubterm(1).termString(); 
    }

    public String toString() {
	return NODE_STRING;
    }

    public String elementName() {
	return FUI_XML.POWER_ELEM;
    }

    public Element toXML() {
	
	Element elem = new Element(elementName());
	elem.addContent(getBase().toXML());
	elem.setAttribute(FUI_XML.EXP_ATTR, getExponent().toString());
	return elem;
    }

    public int events() {
	return getBase().events() * getExponent().value();
    }

    public EventIterator eventIterator() {
	return new PowerIterator();
    }

    class PowerIterator implements EventIterator {

	private int current;
	private EventIterator itr;

	public PowerIterator() {
	    current = 1;
	    itr = getBase().eventIterator();
	}

	public boolean hasNext() {
	    return itr.hasNext();
	}

	public AtomicEvent next() throws NoSuchElementException {

	    AtomicEvent n = itr.next();

	    if (!itr.hasNext() && current < getExponent().value()) {
		current++;
		itr = getBase().eventIterator();
	    }

	    return n;
	}
    }

    public Term copyTerm(EventAlgebra a) {

	EventTerm e = (EventTerm) getBase().copyTerm(a);
	IntTerm i = (IntTerm) getExponent().copyTerm(a);
	return new Power(e, i, a);
    }
}

