package org.uclic.fui;

import org.uclic.util.ArrayUtils;

/**
 * A finite state machine (FSM).  This is defined as:
 * <ul>
 *   <li>a set of states, with a distinguished initial state
 *   <li>a set of events
 *   <li>a transition relation (state * events * state)
 * </ul>
 *
 * <p>
 * The class allows states and events to be added, renamed and
 * removed.  Transitions can be accessed and defined.
 * </p>
 *
 * @version 0.4
 * @author Jeremy Gow
 */
public class FSM {

    // *************** FIELDS ***************

    // List of states.
    protected String[]   states;

    // The initial state.
    protected int        initial;
    
    // List of events.
    protected String[]   events;
  
    /* Transition table.  trans[state][event] is an unsorted
     * array of range states.
     */ 
    protected int[][][]  trans;
    
    // List of default initial states.
    protected final static String[] INIT_STATES = {"StateA"};

    // List of default initial events.
    protected final static String[] INIT_EVENTS = {"Event1"};


    // *************** CONSTRUCTORS ***************

    /* Creates a minimal FSM, with one state and one event.  No
     * transitions are created.
     */
    public FSM() {
        this(INIT_STATES, INIT_EVENTS);
    }

    /* Creates an FSM with the given states and events.  No transitions
     * are created.
     */
    public FSM(String[] s, String[] e) {
	states = s;
	initial = 0;
	events = e;
	trans = new int[states.length][events.length][];
    }

    public FSM(FSM fsm) {
	
	String[] s = fsm.states();
	states = new String[s.length];
	for (int i = 0; i < states.length; i++) {
	    states[i] = new String(s[i]);
	}

	String[] e = fsm.events();
	events = new String[e.length];
	for (int i = 0; i < events.length; i++) {
	    events[i] = new String(e[i]);
	}
	
	initial = fsm.initialStateIndex();

	trans = new int[states.length][events.length][];
	for (int i = 0; i < states.length; i++) {
	    for (int j = 0; j < events.length; j++) {
		int[] k = fsm.rangeIndices(i, j);
		if (k != null) {
		    trans[i][j] = (int[]) k.clone();
		}
	    }
	}
    }


    // *************** NAMESPACE METHODS ***************

    public boolean isName(String name) {
	return isState(name) || isEvent(name);
    }

    public class NamespaceException extends FUIException {
	
	public NamespaceException(String name) {
	    super("The name '" + name + "' is already in use.");
	}
    }
    

    // *************** STATE METHODS ***************
    
    /* The list of states.
     */
    public String[] states() {
        return states;
    }
    
    /* Is this a state?
     */
    public boolean isState(String name) {
	return stateIndex(name) != -1;
    }
    
    /* The index of a given state.
     */
    public int stateIndex(String name) {
	return ArrayUtils.findIndex(name, states);
    }
    
    /* Is this a state index?
     */
    public boolean isStateIndex(int i) {
	return (i > -1 && i < states.length);
    }
    
    /* Returns the indexed state.  An exception occurs if the index is
     * out of bounds.
     */
    public String state(int i) throws FUIException {
	if (isStateIndex(i)) {
	    return states[i];
	} else {
	    throw new FUIException("State index out of bounds: " + i);
	}
    }

    /* Adds a state to the FSM.  An exception occurs if the name is
     * already in use.  Subclasses should override addStateToMachine
     * instead of this method.
     */
    public void addState(String name) throws FUIException {
	if (!isName(name)) {
	    addStateToMachine(name);
	} else {
	    throw new NamespaceException(name);
	}
    }

    /* Adds a state to the FSM - subclasses should override this
     * method instead of addState.
     */
    protected void addStateToMachine(String name) {
	states = ArrayUtils.addToArray(name, states);
	trans = ArrayUtils.addRowToArrayTable(trans);
    }

    /* Deletes a state from the FSM.  No action is taken if the index
     * is out of bounds, or if only one state remains.  Subclasses
     * should override deleteStateFromMachine instead of this method.
     */
    public void deleteState(int i) {
	if (isStateIndex(i) && states.length > 1) {
	    deleteStateFromMachine(i);
	}
    }

    /* Deletes a state from the FSM - subclasses should override this
     * method instead of deleteState.
     */
    protected void deleteStateFromMachine(int i) {
	states = ArrayUtils.deleteFromArray(i, states);
	trans = ArrayUtils.deleteRowFromArrayTable(i, trans);
	trans = ArrayUtils.deleteAllFromArrayTable(i, trans);
    }

    /* Rename the indexed state.  No action is taken if the index is
     * out of bounds.  An exception occurs if the name is already in
     * use.
     */
    public void renameState(int i, String name) throws FUIException {
	if (isStateIndex(i) && !name.equals(states[i])) {
	    if (!isName(name)) {
		states[i] = name;
	    } else {
		throw new NamespaceException(name);
	    }
	}
    }

	
    // *************** INITIAL STATE METHODS ***************

    /* The initial state.
     */
    public String initialState() {
	return states[initial];
    }

    /* The index of the initial state.
     */
    public int initialStateIndex() {
	return initial;
    }

    /* Set the initial state.  No action is taken if the state does
     * not exist.
     */
    public void setInitialState(String name) {

	int i = stateIndex(name);
	
	if (i != -1) {
	    initial = i;
	}
    }

    /* Set the initial state to the indexed state.  No action is taken
     * if the index is out of bounds.
     */
    public void setInitialState(int i) {
	if (isStateIndex(i)) {
	    initial = i;
	}
    }


    // *************** EVENTS METHODS ***************

    
    /* The list of events.
     */
    public String[] events() {
	return events;
    }

    /* Is this an event?
     */
    public boolean isEvent(String name) {
	return eventIndex(name) != -1;
    }

    /* The index of given event.
     */
    public int eventIndex(String name) {
	return ArrayUtils.findIndex(name, events);
    }

    /* Is this is an event index?
     */
    public boolean isEventIndex(int i) {
	return (i > -1 && i < events.length);
    }

    /* Returns the indexed event.  An exception occurs if the index is
     * out of bounds.
     */
    public String event(int i) throws FUIException {
	if (isEventIndex(i)) {
	    return events[i];
	} else {
	    throw new FUIException("Event index out of bounds: " + i);
	}
    }
    
    /* Adds an event to the FSM.  An exception occurs if the name is
     * already in use.  Subclasses should override addEventToMachine
     * instead of this method.
     */
    public void addEvent(String name) throws FUIException {

	if (!isName(name)) {
	    addEventToMachine(name);
	} else {
	    throw new NamespaceException(name);
	}
    }

    /* Adds an event to the FSM - subclasses should override this
     * method instead of addEvent.
     */
    protected void addEventToMachine(String name) {
	events = ArrayUtils.addToArray(name, events);
	trans = ArrayUtils.addColumnToArrayTable(trans);
    }

    /* Deletes an event from the FSM.  No action is taken if the index
     * is out of bounds, or if only one event remains.  Subclasses
     * should override deleteEventFromMachine instead of this method.
     */
    public void deleteEvent(int i) {
	if (isEventIndex(i) && events.length > 1) {
	    deleteEventFromMachine(i);
	}
    }

    /* Deletes an event from the FSM - subclasses should override this
     * method instead of deleteEvent.
     */
    protected void deleteEventFromMachine(int i) {
	events = ArrayUtils.deleteFromArray(i, events);
	trans = ArrayUtils.deleteColumnFromArrayTable(i, trans); 
    } 

    /* Rename the indexed event.  No action is taken if the index is
     * out of bounds.  An exception occurs if the name is already in
     * use.
     */
    public void renameEvent(int i, String name) throws FUIException {
	if (isEventIndex(i) && !name.equals(events[i])) {
	    if (!isName(name)) {
		events[i] = name;
	    } else {
		throw new NamespaceException(name);
	    }
	}
    }


    // *************** RANGE METHODS ***************

    public boolean[] range(int domain, int event) {
        return ArrayUtils.toPredicate(trans[domain][event], states.length);
    }

    public int[] rangeIndices(int domain, int event) {
        return trans[domain][event];
    }

    // Deprecate?  Too data-dependant.
    public int[][] rangeIndices(int state) {
        return trans[state];
    }    
    
    /*
    public int[][][] rangeIndices() {
	return trans;
    }
    */
 
    // *************** DOMAIN METHODS ***************

    public boolean[] domain(int event, int state) {
	
	boolean[] domain = new boolean[states.length];
	
	for (int i = 0; i < domain.length; i++) {
	    if (ArrayUtils.inArray(state, rangeIndices(i, event))) {
		domain[i] = true;
	    }
	}
	return domain;
    }

    public int[] domainIndices(int event, int state) {
	return ArrayUtils.toSet(domain(event, state));
    }


    // *************** MAP METHODS ***************

    public boolean[] maps(int domain, int range) {

	boolean[] maps = new boolean[events.length];

	for (int i = 0; i < maps.length; i++) {
	    if (ArrayUtils.inArray(range, rangeIndices(domain, i))) {
		maps[i] = true;
	    }
	}
	return maps;
    }

    public int[] mapsIndices(int domain, int range) {
	return ArrayUtils.toSet(maps(domain, range));
    }


    // *************** TRANSITION METHODS ***************

    /* Is this a transition?
     */
    public boolean isTransition(int domain, int event, int range) {
	if (isStateIndex(domain) && isEventIndex(event) && isStateIndex(range)) {
	    return ArrayUtils.isArrayTableEntry(domain, event, range, trans);
	} else {
	    return false;
	}
    }

    public void toggleTransition(int domain, int event, int range) {
	if (isStateIndex(domain) && isEventIndex(event) && isStateIndex(range)) {
	    if (ArrayUtils.isArrayTableEntry(domain, event, range, trans)) {
		deleteTransFromMachine(domain, event, range);
	    } else {
		addTransToMachine(domain, event, range);
	    }
	} 
    }

    /* Adds a transition to the FSM.  Subclasses should override
     * addTransToMachine instead of this method.
     */
    public void addTransition(int domain, int event, int range) {
	
	if (isStateIndex(domain) && isEventIndex(event) && isStateIndex(range)
	    && !isTransition(domain, event, range)) {

	    addTransToMachine(domain, event, range);
	}
    }

    /* Adds a transition from the FSM - subclasses should override this
     * method instead of addTransition.
     */
    protected void addTransToMachine(int domain, int event, int range) {
	trans = ArrayUtils.addArrayTableEntry(domain, event, range, trans);
    }	
	
    /* Deletes a transition from the FSM.  No action is taken if any
     * of the indices is out of bounds.  Subclasses should override
     * deleteTransFromMachine instead of this method.
     */
    public void deleteTransition(int domain, int event, int range) {
	
	if (isStateIndex(domain) && isEventIndex(event) && isStateIndex(range)
	    && isTransition(domain, event, range)) {
	    
	    deleteTransFromMachine(domain, event, range);
	}
    }

    /* Deletes a transition from the FSM - subclasses should override this
     * method instead of deleteTransition.
     */
    protected void deleteTransFromMachine(int domain, int event, int range) {
	trans = ArrayUtils.deleteArrayTableEntry(domain, event, range, trans);
    }

    /* Deletes all the transitions out of a given state for a given
     * event.  No action is taken if either index is out of bounds.
     * Subclasses should override deleteTransFromMachine instead of
     * this method.
     */
    public void deleteTransitions(int domain, int event) {
	
	if (isStateIndex(domain) && isEventIndex(event)) {
	    
	    int[] range = trans[domain][event];

	    for (int i = 0; i < range.length; i++) {
		deleteTransFromMachine(domain, event, range[i]);
	    }
	}
    } 
   
    /* Deletes all the transitions out of a given state.  No action is
     * taken if the index is out of bounds.  Subclasses should
     * override deleteTransFromMachine instead of this method.
     */
    public void deleteTransitions(int domain) {

	if (isStateIndex(domain)) {
	    
	    for (int event = 0; event < events.length; event++) {

		int[] range = trans[domain][event];

		for (int i = 0; i < range.length; i++) {
		    deleteTransFromMachine(domain, event, range[i]);
		}
	    }
	}
    }

       
}
