package org.uclic.fui;

import java.util.ListIterator;
import java.util.Vector;
import org.uclic.algebra.*;


/**
 * A formal interface design. Consisting of:
 * <ul>
 * <li>a name;</li>
 * <li>an extended FSM model of the design;</li>
 * <li>the associated event algebra;</li>
 * <li>an algebraic specification.</li>   
 * </ul>
 *
 * @author Jeremy Gow
 * @version 0.4
 */
public class FUIDesign {

    private String        name;
    private SignFSM       model;
    private EventAlgebra  algebra;
    private Specification spec;

    private SignFSM       reserveModel;
    private EventAlgebra  reserveAlgebra;
    private Specification reserveSpec;

    private Vector        skinSrc;
    private Vector        skinType;
    
    public FUIDesign(String n, SignFSM m) {
	super();
	name = n;
	setModel(m);
	skinSrc = new Vector();
	skinType = new Vector();
    }

    public FUIDesign() {
	this("New Design", new SignFSM());
    }

    /**
     * Returns the name of the interface design.
     *
     * @return a string of the name.
     */
    public String name() {
	return name;
    }

    /**
     * Set the name of the interface design.
     *
     * @param n a string of the new name.
     * @throws FUIException If the string is empty. 
     */
    public void setName(String n) throws FUIException {
        if (!n.equals("")) {
	    name = n;
	} else {
	    throw new FUIException("Design name can't be empty.");
	}
    }


    /**
     * Returns the finite state machine model.
     *
     * @return the finite state machine.
     */
    public SignFSM model() {
	return model;
    }

    public void setModel(SignFSM m) {
	model = m;
	algebra = new EventAlgebra(model);
	spec = new Specification(algebra);
    }

    public void reserveModel() {
	reserveModel = new SignFSM(model);
	reserveAlgebra = new EventAlgebra(reserveModel);
	//reserveSpec = new Specification(reserveAlgebra, spec);
	    
    }

    public void restoreModel() {
	if (reserveModel != null) {
	    model = reserveModel;
	    algebra = reserveAlgebra;
	    spec.setAlgebra(algebra);
	    //spec = reserveSpec;
	    reserveModel = null;
	    reserveAlgebra = null;
	    reserveSpec = null;
	}
    }

    /**
     * Returns the event algebra.
     *
     * @return the event algebra.
     */
    public EventAlgebra algebra() {
	return algebra;
    }

    /**
     * Skins
     */
    public String getSkin(int i) {
	return (String) skinSrc.get(i);
    }

    public String getSkinType(int i) {
	return (String) skinType.get(i);
    }

    public Vector getSkins() {
	return skinSrc;
    }
    public Vector getSkinTypes() {
	return skinType;
    }

    public void removeSkin(int i) {
	skinSrc.remove(i);
	skinType.remove(i);
    }

    public void removeSkins() {
	skinSrc.removeAllElements();
	skinType.removeAllElements();
    }

    public void addSkin(String src, String type) {
	skinSrc.add(src);
	skinType.add(type);
    }


    /**
     * Returns the event algebra specification.
     *
     * @return the specification.
     */
    public Specification spec() {
	return spec;
    }

    public void setSpec(Specification s) {
	spec = s;
	spec.setAlgebra(algebra);
    }


    public void deleteEvent(int i) {
	model.deleteEvent(i);
	algebra.deleteEvent(i);
	spec.deleteEvent(i);
    }

    public void deleteState(int i) {
	model.deleteState(i);
	algebra.deleteState(i);
	spec.deleteState(i);
    }

    public void deleteSuperstate(int i) {
	model.deleteSuperstate(i);
	algebra.deleteSuperstate(i);
	spec.deleteSuperstate(i);
    }

    public void deleteSign(int i) {
	model.deleteSign(i);
	algebra.deleteSign(i);
	spec.deleteSign(i);
    }

}
