package org.uclic.fui;

import java.util.*;
import org.jdom.Element;
import org.uclic.algebra.*;
import org.uclic.algebra.property.*;
import org.uclic.algebra.event.*;
import org.uclic.algebra.state.*;

/**
 * XML reader for an interface model.
 *
 * @version 0.4
 * @author Jeremy Gow
 */
public class FUIReader {

    private Element root;

    private String[] states;
    private String[] events;
    private String[] superstates;
    private String[] signs;

    private Hashtable stateTable;
    private Hashtable eventTable;
    private Hashtable superTable;
    private Hashtable signTable;

    private Specification spec;
    private EventAlgebra  alg;

    public FUIReader(Element elem) {
	root = elem;
	stateTable = new Hashtable();
	eventTable = new Hashtable();
	superTable = new Hashtable();
	signTable = new Hashtable();
    }

    private int getIndex(Hashtable table, String type, String name) throws FUIException {
	Integer i = (Integer) table.get(name);
	if (i != null) {
	    return i.intValue();
	} else {
	    throw new FUIException(type + " '" + name + "' does not exist.");
	}
    }

    private int getStateIndex(String name) throws FUIException {
	return getIndex(stateTable, "State", name);
    }

    private int getEventIndex(String name) throws FUIException {
	return getIndex(eventTable, "Event", name);
    }

    private int getSuperIndex(String name) throws FUIException {
	return getIndex(superTable, "Superstate", name);
    }

    private int getSignIndex(String name) throws FUIException {
	return getIndex(signTable, "Sign", name);
    }


    /**
     * Reads a FUI design from XML.
     *
     * @param root A <fui> org.jdom.Element representing a FUI design.
     */
    public FUIDesign read() throws FUIException {

	if (!root.getName().equals(FUI_XML.ROOT_ELEM)) {
	    throw new FUIException("'fui' element expected, not '" + root.getName() + "'");
	}

	// Read in the name
	Element nameElem = root.getChild(FUI_XML.NAME_ELEM);
	String name = nameElem.getTextTrim();

	// Read in the model
	SignFSM model = readModel();

	// Construct a design object
	FUIDesign design = new FUIDesign(name, model);
	spec = design.spec();
	alg = spec.algebra();

	// Read in skins
       	Element form = (Element) root.getChild(FUI_XML.FORM_ELEM);
	if (form != null) {
	    
	    ListIterator skins = form.getChildren(FUI_XML.SKIN_ELEM).listIterator();
	    Element skin;
	    String src, type;

	    while (skins.hasNext()) {
		skin = (Element) skins.next();
		src = skin.getAttributeValue(FUI_XML.HREF_ATTR);
		type = skin.getAttributeValue(FUI_XML.TYPE_ATTR);
		design.addSkin(src, type);
	    }
	}
	
	// Read in the specification
	readSpecification();

	return design;
    }



    
    /**
     * Reads a FUI model from XML.
     *
     * @param elem A <func> org.jdom.Element representing a FUI model.
     */
    private SignFSM readModel() throws FUIException {

       	Element form, func, elem, subElem;
	String id, ref;
	int ste, cls, evt, to, sgn;
	ListIterator itr, subItr;
	
	form = (Element) root.getChild(FUI_XML.FORM_ELEM);
	func = (Element) root.getChild(FUI_XML.FUNC_ELEM);

	// Read states and events
	states = readElements(func, FUI_XML.STATE_ELEM, stateTable);
	events = readElements(func, FUI_XML.EVENT_ELEM, eventTable);
	superstates = readElements(func, FUI_XML.SUPER_DEF_ELEM, superTable);

	signs = readElements(form, FUI_XML.SIGN_ELEM, signTable);
	
	// Construct model object
	SignFSM model = new SignFSM(states, events, superstates, signs);

	// Read initial
	elem = func.getChild(FUI_XML.INITIAL_ELEM);
	if (elem != null) {
	    ref = elem.getAttributeValue(FUI_XML.REF_ATTR);	
	    model.setInitialState(getStateIndex(ref));
	} else {
	    throw new FUIException("Initial state not given.");
	}

	readSuperData(func, model);
	readStateData(func, model);

	return model;
    }


    private void readSuperData(Element func, SignFSM model) throws FUIException {

       	Element elem, subElem;
	int ste, cls, evt, to, sgn;
	ListIterator itr, subItr;

	// Read superstate elements for transitions and signs
	itr = func.getChildren(FUI_XML.SUPER_DEF_ELEM).listIterator();
	while (itr.hasNext()) {
	    // Read superstate element
	    elem = (Element) itr.next();
	    cls = getSuperIndex(elem.getAttributeValue(FUI_XML.ID_ATTR));

	    // Read transitions for this superstate
	    subItr = elem.getChildren(FUI_XML.CHANGE_ELEM).listIterator();	    
	    while (subItr.hasNext()) {
		subElem = (Element) subItr.next();		
		evt = getEventIndex(subElem.getAttributeValue(FUI_XML.EVENT_ATTR));
		to = getStateIndex(subElem.getAttributeValue(FUI_XML.TO_ATTR));
		// Record transition for superstate
		model.setSuperTransition(cls, evt, to, true);
	    }
	    	    
	    // Read identity transitions for this superstate
	    subItr = elem.getChildren(FUI_XML.NOCHANGE_ELEM).listIterator();	    
	    while (subItr.hasNext()) {
		subElem = (Element) subItr.next();		
		evt = getEventIndex(subElem.getAttributeValue(FUI_XML.EVENT_ATTR));
		// Record identity transition for superstate
		model.setSuperIdentity(cls, evt, true);
	    }

	    // Read signs for this superstate
	    subItr = elem.getChildren(FUI_XML.DISPLAY_ELEM).listIterator();
	    while (subItr.hasNext()) {
		subElem = (Element) subItr.next();
		sgn = getSignIndex(subElem.getAttributeValue(FUI_XML.REF_ATTR));
		// Record sign for superstate
		model.setSuperstateSign(cls, sgn, true);
	    }
	}
    }

    private void readStateData(Element func, SignFSM model) throws FUIException {
	
       	Element elem, subElem;
	int ste, cls, evt, to, sgn;
	ListIterator itr, subItr;

	// Read state elements for superstate membership, transitions and signs
	itr = func.getChildren(FUI_XML.STATE_ELEM).listIterator();
	while (itr.hasNext()) {
	    // Read state element
	    elem = (Element) itr.next();
	    ste = getStateIndex(elem.getAttributeValue(FUI_XML.ID_ATTR));
	    
	    // Read superstate membership for this state
	    subItr = elem.getChildren(FUI_XML.SUPER_REF_ELEM).listIterator();	    
	    while (subItr.hasNext()) {
		subElem = (Element) subItr.next();		
		cls = getSuperIndex(subElem.getAttributeValue(FUI_XML.REF_ATTR));
		// Record superstate member
		model.setIncluded(ste, cls, true);
	    }
	    	    
	    // Read transitions for this state
	    subItr = elem.getChildren(FUI_XML.CHANGE_ELEM).listIterator();	    
	    while (subItr.hasNext()) {
		subElem = (Element) subItr.next();		
		evt = getEventIndex(subElem.getAttributeValue(FUI_XML.EVENT_ATTR));
		to = getStateIndex(subElem.getAttributeValue(FUI_XML.TO_ATTR));
		// Record transition for state
		model.addTransition(ste, evt, to);
	    }

	    // Read identity transitions for this state
	    subItr = elem.getChildren(FUI_XML.NOCHANGE_ELEM).listIterator();	    
	    while (subItr.hasNext()) {
		subElem = (Element) subItr.next();		
		evt = getEventIndex(subElem.getAttributeValue(FUI_XML.EVENT_ATTR));
		// Record identity transition for state
		model.addTransition(ste, evt, ste);
	    }

	    // Read signs for this state
	    subItr = elem.getChildren(FUI_XML.DISPLAY_ELEM).listIterator();
	    while (subItr.hasNext()) {
		subElem = (Element) subItr.next();
		sgn = getSignIndex(subElem.getAttributeValue(FUI_XML.REF_ATTR));
		// Record sign for state
		model.setStateSign(ste, sgn, true);
	    }
	}
    }

    private String[] readElements(Element elem, String name, Hashtable table) throws FUIException {

	List elems = elem.getChildren(name);
	String[] array = new String[elems.size()];	
	ListIterator itr = elems.listIterator();
	
	while (itr.hasNext()) {
	    Element element = (Element) itr.next();
	    String id = element.getAttributeValue(FUI_XML.ID_ATTR);
	    idCheck(id);
	    array[itr.previousIndex()] = id;
	    table.put(id, new Integer(itr.previousIndex())); 
	}

	return array;
    }

    private void idCheck(String id) throws FUIException {

	if (id.equals("")) {
	    throw new FUIException("Empty identifier");
	} else {
	    char[] c = id.toCharArray();
	    for (int i = 0; i < c.length; i++) {
		if (c[i] == ':') {
		    throw new FUIException("Identifier '" + id + "' contains illegal character '" + c[i] + "'");
		}
	    }
	}
    }

    
    private void readSpecification() throws FUIException {
	
	Element prop, sp, kid;
	ListIterator props;

	// Read properties
	sp = root.getChild(FUI_XML.SPEC_ELEM);	
	if (sp != null) {
	    props = sp.getChildren(FUI_XML.PROP_ELEM).listIterator();
	    while(props.hasNext()) {
		prop = (Element) props.next();
		kid = (Element) prop.getChildren().get(0);
		spec.addProperty(readProperty(kid));
	    }
	}
    }

    private PropertyTerm readProperty(Element prop) throws FUIException {

	PropertyTerm term;
	String name = prop.getName();

	Element kid1 = (Element) prop.getChildren().get(0);
	
	if (name.equals(FUI_XML.NOT_ELEM)) {
	    // NOT
	    term = new Not(readProperty(kid1), alg);

	} else if (name.equals(FUI_XML.UNDO_ELEM)) {
	    // UNDO
	    term = new Undo(readEvent(kid1), alg);

	} else {
	    // EQUALS
	    EventTerm arg1 = readEvent(kid1);
	    Element kid2 = (Element) prop.getChildren().get(1);

	    if (arg1 != null) {
		// EVENT EQUALITY
		EventTerm arg2 = readEvent(kid2);
		term = new EventEquality(arg1, arg2, alg);

	    } else {
		// STATE EQUALITY
		StateTerm arg3 = readState(kid1);
		StateTerm arg4 = readState(kid2);
		term = new StateEquality(arg3, arg4, alg);
	    }
	}

    	return term;
    }
    
    private EventTerm readEvent(Element event) throws FUIException {

	EventTerm term;
	String name = event.getName();
	
	if (name.equals(FUI_XML.UIEVENT_ELEM)) {
	    // ATOMIC EVENT
	    String ref = event.getAttributeValue(FUI_XML.REF_ATTR);	
	    term = new UIEvent(getEventIndex(ref), alg);

	} else if (name.equals(FUI_XML.ID_ELEM)) {
	    // ID
	    term = new Id(alg);
	
	} else {
	    // AT LEAST ONE ARGUMENT
	    List kids = event.getChildren();
	    Element kid1 = (Element) kids.get(0);

	    if (name.equals(FUI_XML.GOTO_ELEM)) {
		// GOTO
		term = new Goto(readState(kid1), alg);
	    
	    } else if (name.equals(FUI_XML.POWER_ELEM)) {
		// POWER
		String times = event.getAttributeValue(FUI_XML.EXP_ATTR);
		IntTerm exp = new IntTerm(Integer.parseInt(times), alg);
		term = new Power(readEvent(kid1), exp, alg);

	    } else {
		// MULTIPLE ARGUMENTS
		ListIterator itr = kids.listIterator();
		itr.next(); // Pop kid1

		if (name.equals(FUI_XML.EPROD_ELEM)) {
		    // EVENT PRODUCT
		    term = new EventProduct(readEvent(kid1), alg);
		} else {
		    // EVENT SUM
		    term = new EventSum(readEvent(kid1), alg);
		}
		while (itr.hasNext()) {
		    term.add(readEvent((Element) itr.next()));
		}
	    }
	}
	return term;
    }


    private StateTerm readState(Element state) throws FUIException {

	StateTerm term;
	String name = state.getName();
	
	if (name.equals(FUI_XML.UISTATE_ELEM)) {
	    // ATOMIC STATE
	    String ref = state.getAttributeValue(FUI_XML.REF_ATTR);	
	    term = new UIState(getStateIndex(ref), alg);

	} else if (name.equals(FUI_XML.SUPER_ELEM)) {
	    // ATOMIC SUPERSTATE
	    String ref = state.getAttributeValue(FUI_XML.REF_ATTR);	
	    term = new UISuper(getSuperIndex(ref), alg);

	} else {
	    // AT LEAST ONE ARGUMENT
	    List kids = state.getChildren();
	    Element kid1 = (Element) kids.get(0);
	    
	    if (name.equals(FUI_XML.SPROD_ELEM)) {
		// TRANSITION
		Element kid2 = (Element) kids.get(1);
		term = new Transition(readState(kid1), readEvent(kid2), alg);

	    } else {
		// STATE SUM
		term = new StateSum(readState(kid1), alg);
		ListIterator itr = kids.listIterator();
		while (itr.hasNext()) {
		    term.add(readState((Element) itr.next()));
		}
	    }
	}
	return term;
    }

}

