package org.uclic.fui;

import java.util.*;
import org.jdom.Element;

/**
 * XML writer for an interface model.
 *
 * @author Jeremy Gow
 * @version 0.4
 */
public class FUIWriter {

    private FUIDesign  design;
    private SignFSM    model;
    private Element    fui;
    private Element    func;

    public FUIWriter(FUIDesign d) {
	design = d;
	model = design.model();
    }
    
    /**
     * Writes the design to XML.
     *
     * @return A FUI org.jdom.Element representing the design.
     */
    public Element write() {

	Element name;

	fui = new Element(FUI_XML.ROOT_ELEM);
	
	// Write name
	name = new Element(FUI_XML.NAME_ELEM);
	name.addContent(design.name());
	fui.addContent(name);

	writeForm();
	writeFunction();
	
	if (!design.spec().isEmpty()) {
	    fui.addContent(design.spec().toXML());
	}

	return fui;
    }

    /**
     * Write design form to XML.
     *
     */
    protected void writeForm() {

	Element form, sign;
	String[] signs;

	form = new Element(FUI_XML.FORM_ELEM);
	fui.addContent(form);

	// Write skin elements
	int s = design.getSkins().size();
	for (int i = 0; i < s; i++) {
	    Element skin = new Element(FUI_XML.SKIN_ELEM);
	    skin.setAttribute(FUI_XML.HREF_ATTR, design.getSkin(i));
	    skin.setAttribute(FUI_XML.TYPE_ATTR, design.getSkinType(i));
	    form.addContent(skin);
	}
	
	// Write sign elements
	signs = model.signs();
	for (int i = 0; i < signs.length; i++) {
	    sign = new Element(FUI_XML.SIGN_ELEM);
	    sign.setAttribute(FUI_XML.ID_ATTR, signs[i]);
	    form.addContent(sign);
	}
    }

    /**
     * Write design function to XML.
     *
     */
    protected void writeFunction() {

	func = new Element(FUI_XML.FUNC_ELEM);
	fui.addContent(func);
	
	try {
	    writeEvents();
	    writeInitial();
	    writeSuperstates();
	    writeStates();
	} catch (FUIException e) {
	    func.addContent("ERROR");
	}
    }

    /**
     * Write events to XML.
     *
     */
    protected void writeEvents() throws FUIException {

	for(int i = 0; i < model.events().length; i++) {
	    Element event = new Element(FUI_XML.EVENT_ELEM);
	    event.setAttribute(FUI_XML.ID_ATTR, model.event(i));
	    func.addContent(event);

	    /*
	    try {
		Element matrix = design.algebra().eventMatrix(i).toXML();
		func.addContent(matrix);
	    } catch (org.uclic.algebra.EvaluationException e) {}
	    */
	}
    }

    /**
     * Write initial element to XML.
     *
     */
    protected void writeInitial() {

	Element initElem = new Element(FUI_XML.INITIAL_ELEM);
	String initial =  model.initialState();
	initElem.setAttribute(FUI_XML.REF_ATTR, initial);
	func.addContent(initElem);
    }	


    /**
     * Write state superstates to XML.
     *
     */
    protected void writeSuperstates() throws FUIException {

	// Write superstate definitions
	for (int sup = 0; sup < model.superstates().length; sup++) {
	    Element superElem = new Element(FUI_XML.SUPER_DEF_ELEM);
	    superElem.setAttribute(FUI_XML.ID_ATTR, model.superstate(sup));
	    func.addContent(superElem);
	    
	    // Write superstate changes
	    int[][] row = model.superRangeIndices(sup);
	    
	    for(int evt = 0; evt < row.length; evt++) {
		
		int [] entry = row[evt];
		
		if (entry != null) {
		    for (int e = 0; e < entry.length; e++) {
			Element change = new Element(FUI_XML.CHANGE_ELEM);
			superElem.addContent(change);
			change.setAttribute(FUI_XML.EVENT_ATTR, model.event(evt));
			change.setAttribute(FUI_XML.TO_ATTR, model.state(entry[e]));
		    }
		}
	    }
	    
	    // Write superstate signs
	    for (int sgn = 0; sgn < model.signs().length; sgn++) {
		if (model.isSuperstateSign(sup, sgn)) {
		    Element sign = new Element(FUI_XML.DISPLAY_ELEM);
		    superElem.addContent(sign);
		    sign.setAttribute(FUI_XML.REF_ATTR, model.sign(sgn));
		}
	    }
	}
    }

    /**
     * Write states to XML.
     *
     */
    protected void writeStates() throws FUIException {

	// Write state definitions
	for (int ste = 0; ste < model.states().length; ste++) {
	    Element stateElem = new Element(FUI_XML.STATE_ELEM);
	    stateElem.setAttribute(FUI_XML.ID_ATTR, model.state(ste));
	    func.addContent(stateElem);
	    int[] belongs = model.superstatesIndices(ste);

	    // Write superstate membership
	    if (belongs != null) {
		for (int b = 0; b < belongs.length; b++) {
		    Element superElem = new Element(FUI_XML.SUPER_REF_ELEM);
		    superElem.setAttribute(FUI_XML.REF_ATTR, model.superstate(belongs[b]));
		    stateElem.addContent(superElem);
		}
	    }
	    
	    // Write state changes
	    for (int evt = 0; evt < model.events().length; evt++) {		
		if (model.range(ste, evt) != null) {
		    // Transition: ste --evt--> to
		    for (int k = 0; k < model.rangeIndices(ste, evt).length; k++) {
			
			int to =  model.rangeIndices(ste, evt)[k];

			// Check if transition defined by some superstate
			boolean already = false;
			if (belongs != null) {
			    for (int b = 0; b < belongs.length; b++) {
				if (model.isSuperTransition(belongs[b], evt, to)
				    || (ste == to && model.isSuperIdentity(belongs[b], evt))) {
				    already = true;
				    break;
				}
			    }
			}

			// Write transition if it is not already defined
			if (!already) {
			    Element change = new Element(FUI_XML.CHANGE_ELEM);
			    change.setAttribute(FUI_XML.EVENT_ATTR, model.event(evt));
			    change.setAttribute(FUI_XML.TO_ATTR, model.state(model.rangeIndices(ste, evt)[k]));
			    stateElem.addContent(change);
			}
		    }
		}
	    }
	    
	    // Write state signs
	    for (int sgn = 0; sgn < model.signs().length; sgn++) {
		if (model.isStateSign(ste, sgn)) {
		    
		    // Check if sign already defined by some superstate
		    boolean already = false;
		    if (belongs != null) {
			for (int b = 0; b < belongs.length; b++) {
			    if (model.isSuperstateSign(belongs[b], sgn)) {
				already = true;
				break;
			    }
			}
		    }
		    
		    // Write sign if not already defined
		    if (!already) {
			Element signElem = new Element(FUI_XML.DISPLAY_ELEM);
			signElem.setAttribute(FUI_XML.REF_ATTR, model.sign(sgn));
			stateElem.addContent(signElem);
		    }
		}
	    }
	}
    }

}
