package org.uclic.fui;

import java.util.*;
import org.uclic.matrix.*;
import org.uclic.util.ArrayUtils;

/**
 * A finite state/superstate machine with observable signs.  Signs are defined by:
 * <ul>
 * <li>a set of observable signs (sign);</li>
 * <li>a state sign relation (state * sign);</li>
 * <li>a superstate sign relation (superstate * sign).</li>
 * </ul>
 * <p>
 * Signs may be defined, renamed and deleted.
 * </p>
 *
 * @version 0.4
 * @author Jeremy Gow
 */
public class SignFSM extends SuperFSM {

    // *************** FIELDS ***************

    protected String[]    signs;
    protected boolean[][] stateSign;
    protected boolean[][] superSign;

    protected final static String[] INIT_SIGNS = {"Sign"};


    // *************** CONSTRUCTORS ***************

    public SignFSM() {
	this(FSM.INIT_STATES, FSM.INIT_EVENTS, SuperFSM.INIT_SUPER, INIT_SIGNS);
    }

    public SignFSM(String[] s, String[] e, String[] c, String[] g) {
	super(s, e, c);

	// Sign data
	signs = g;
	stateSign = new boolean[states.length][signs.length];
	superSign = new boolean[superstates.length][signs.length];
    }

    public SignFSM(SignFSM fsm) {
	super(fsm);
	
       	String[] s = fsm.signs();
	signs = new String[s.length];
	for (int i = 0; i < signs.length; i++) {
	    signs[i] = new String(s[i]);
	}

	stateSign = new boolean[states.length][signs.length];
	for (int i = 0; i < states.length; i++) {
	    stateSign[i] = (boolean[]) fsm.stateSigns(i).clone();
	}

	superSign = new boolean[superstates.length][signs.length];
	for (int i = 0; i < superstates.length; i++) {
	    superSign[i] = (boolean[]) fsm.superstateSigns(i).clone();
	}
    }


    // *************** NAMESPACE METHODS ***************

    public boolean isName(String name) {
	return super.isName(name) || isSign(name);
    }


    // *************** SIGNS METHODS ***************

    /* The list of signs.
     */
    public String[] signs() {
	return signs;
    }
     
    /* Is this a sign?
     */
    public boolean isSign(String name) {
	return signIndex(name) != -1;
    }
    
    /* The index of a given sign.
     */
    public int signIndex(String name) {
	return ArrayUtils.findIndex(name, signs);
    }
    
    /* Is this a sign index?
     */
    public boolean isSignIndex(int i) {
	return (i > -1 && i < signs.length);
    }

    /* Returns the indexed sign.  An exception occurs if the index is
     * out of bounds.
     */
    public String sign(int i) throws FUIException {
	if (isSignIndex(i)) {
	     return signs[i];
	} else {
	    throw new FUIException("Sign index out of bounds: " + i);
	}
    }

    /* Adds a sign to the FSM.  An exception occurs if the name is
     * already in use.  Subclasses should override addSignToMachine
     * instead of this method.
     */
    public void addSign(String name) throws FUIException {
	if (!isName(name)) {
	    addSignToMachine(name);
	} else {
	    throw new NamespaceException(name);
	}
    }

    /* Adds a sign to the FSM - subclasses should override this
     * method instead of addSign.
     */
    protected void addSignToMachine(String name) {
	signs = ArrayUtils.addToArray(name, signs);
	stateSign = ArrayUtils.addColumnToArrayTable(stateSign);
	superSign = ArrayUtils.addColumnToArrayTable(superSign);
    }


    /* Deletes a sign from the FSM.  No action is taken if the index
     * is out of bounds, or if only one sign remains.  Subclasses
     * should override deleteSignFromMachine instead of this method.
     */
    public void deleteSign(int i) {
	if (isSuperstateIndex(i) && signs.length > 1) {
	    deleteSignFromMachine(i);
	}
    }

    /* Deletes a sign from the FSM - subclasses should override this
     * method instead of deleteSign.
     */
    protected void deleteSignFromMachine(int i) {
	signs = ArrayUtils.deleteFromArray(i, signs);
	stateSign = ArrayUtils.deleteColumnFromArrayTable(i, stateSign);
	superSign = ArrayUtils.deleteColumnFromArrayTable(i, superSign);
    }

    /* Rename the indexed sign.  No action is taken if the index is
     * out of bounds.  An exception occurs if the name is already in
     * use.
     */
    public void renameSign(int i, String name) throws FUIException {
	if (isSignIndex(i) && !name.equals(signs[i])) {
	    if (!isName(name)) {
		signs[i] = name;
	    } else {
		throw new NamespaceException(name);
	    }
	}
    }


    // *************** STATE SIGN METHODS ***************
  
    public boolean isStateSign(int s, int g) {
	return stateSign[s][g];
    }

    public boolean[] stateSigns(int state) {
	return stateSign[state];
    }

    public int[] stateSignIndices(int state) {
	return ArrayUtils.trueInRow(state, stateSign);
    }

    public boolean[] signifiedStates(int sign) {
	return ArrayUtils.column(stateSign, sign);
    }

    public int[] signifiedStateIndices(int sign) {
	return ArrayUtils.trueInColumn(sign, stateSign);
    }

    /* Don't remove signs inherited from a superstate.
     */
    public void setStateSign(int s, int g, boolean val) {
	if (isStateIndex(s) && isSignIndex(g)) {
	    setStateSignForMachine(s, g, val);
	}
    }

    protected void setStateSignForMachine(int state, int sign, boolean val) {
	if (!val) {
	    int[] cs = superstatesIndices(state);
	    if (cs != null) {
		for (int i = 0; i < cs.length; i++) {
		    if (superSign[cs[i]][sign]) {
			val = true;
			break;
		    }
		}
	    }
	}
	
	stateSign[state][sign] = val;
    }

    public void toggleStateSign(int state, int sign) {
	if (isStateIndex(state) && isSignIndex(sign)) {
	    if (stateSign[state][sign]) {
		setStateSignForMachine(state, sign, false);
	    } else {
		setStateSignForMachine(state, sign, true);
	    }
	}
    }

    // *************** STATE SIGN METHODS ***************

    public boolean isSuperstateSign(int c, int g) {
	return superSign[c][g];
    }

    public boolean[] superstateSigns(int sup) {
	return superSign[sup];
    }

    public int[] superstateSignIndicess(int sup) {
	return ArrayUtils.trueInRow(sup, superSign);
    }

    public boolean[] signifiedSuperstates(int sign) {
	return ArrayUtils.column(superSign, sign);
    }

    public int[] signifiedSuperstateIndices(int sign) {
	return ArrayUtils.trueInColumn(sign, superSign);
    }

    public void setSuperstateSign(int superstate, int sign, boolean val) {
	if (isSuperstateIndex(superstate) && isSignIndex(sign)) {
	    setSuperstateSignForMachine(superstate, sign, val);
	}
    }

    protected void setSuperstateSignForMachine(int superstate, int sign, boolean val) {
	superSign[superstate][sign] = val;
    }

    public void toggleSuperstateSign(int superstate, int sign) {
	if (isSuperstateIndex(superstate) && isSignIndex(sign)) {
	    if (superSign[superstate][sign]) {
		setSuperstateSignForMachine(superstate, sign, false);
	    } else {
		setSuperstateSignForMachine(superstate, sign, true);
	    }
	}
    }


    // *************** OVERRIDING SUPERFSM METHODS ***************

    public void addStateToMachine(String name) {
	super.addStateToMachine(name);
	stateSign = ArrayUtils.addRowToArrayTable(stateSign);
    }

    public void deleteStateFromMachine(int i) {
	stateSign = ArrayUtils.deleteRowFromArrayTable(i, stateSign);
	super.deleteStateFromMachine(i);
    }

    public void addSuperstateToMachine(String name) {
	super.addSuperstateToMachine(name);
	superSign = ArrayUtils.addRowToArrayTable(superSign);
    }

    public void deleteSuperstateFromMachine(int i) {
	superSign = ArrayUtils.deleteRowFromArrayTable(i, superSign);
	super.deleteSuperstateFromMachine(i);
    }

    protected void setIncludedInSuperstate(int s, int c, boolean val) {
	super.setIncludedInSuperstate(s, c, val);
	if (val) {
	    // Add class signs to state
	    for (int g = 0; g < signs.length; g++) {
		if (isSuperstateSign(c, g)) {
		    setStateSign(s, g, true);
		}
	    }
	}
    }

}
