package org.uclic.fui;

import org.uclic.util.ArrayUtils;

/**
 * A finite state machine with superstates.  Superstates are defined by:
 * <ul>
 * <li>a set of super states (super);</li>
 * <li>a membership relation (state * super) ;</li>
 * <li>a transition relation (super * event * state);</li>
 * <li>an identity transition relation (super * event);</li>
 * </ul>
 * <p>
 * Superstates may be defined, renamed and deleted.
 * </p>
 *
 * @version 0.4
 * @author Jeremy Gow
 */
public class SuperFSM extends FSM {

    // *************** FIELDS ***************

    // List of superstates.
    protected String[]      superstates;

    /* Membership table.  The first dimension is the state index, the
     * second the superstate index.
     */
    protected boolean[][]   inSuper;

    /* Transition table.  superTrans[superstate][event] is an unsorted
     * array of range states.
     */
    protected int[][][]     superTrans;

    /* Identity transition table.  Records superstates (first
     * dimesion) for which all element states have an identity
     * transition for a given event (second dimension).
     */
    protected boolean[][]   superId;

    // List of default initial superstates.
    protected final static String[] INIT_SUPER = {"Superstate"};


    // *************** CONSTRUCTORS ***************

    /* Creates a minimal FSM, with one state, one event and one
     * superstate.  No transitions are created.
     */
    public SuperFSM() {
	this(FSM.INIT_STATES, FSM.INIT_EVENTS, INIT_SUPER);
    }

    /* Creates an FSM with the given states, events and superstates.
     * No transitions are created.
     */
    public SuperFSM(String[] s, String[] e, String[] c) {
	super(s, e);
	superstates = c;
	inSuper = new boolean[states.length][superstates.length];
	superTrans = new int[superstates.length][events.length][];
	superId = new boolean[superstates.length][events.length];
    }

    public SuperFSM(SuperFSM fsm) {
	super(fsm);

       	String[] s = fsm.superstates();
	superstates = new String[s.length];
	for (int i = 0; i < superstates.length; i++) {
	    superstates[i] = new String(s[i]);
	}

	inSuper = new boolean[states.length][superstates.length];
	for (int i = 0; i < states.length; i++) {
	    inSuper[i] = (boolean[]) fsm.superstates(i).clone();
	}

	superTrans = new int[superstates.length][events.length][];
	for (int i = 0; i < superstates.length; i++) {
	    for (int j = 0; j < events.length; j++) {
		int[] k = fsm.superRangeIndices(i, j);
		if (k != null) {
		    superTrans[i][j] = (int[]) k.clone();
		}
	    }
	}

	superId = new boolean[superstates.length][events.length];
	for (int i = 0; i < superstates.length; i++) {
	    superId[i] = (boolean[]) fsm.superIdentities(i).clone();
	}
    }


    // *************** NAMESPACE METHODS ***************

    public boolean isName(String name) {
	return super.isName(name) || isSuperstate(name);
    }


    // *************** SUPERSTATE METHODS ***************

    public String[] superstates() {
	return superstates;
    }

    public boolean isSuperstate(String name) {
	return superstateIndex(name) != -1;
    }
    
    public int superstateIndex(String name) {
	return ArrayUtils.findIndex(name, superstates);
    }

    public boolean isSuperstateIndex(int i) {
	return (i > -1 && i < superstates.length);
    }

    public String superstate(int i) throws FUIException {
	if (isSuperstateIndex(i)) {
	     return superstates[i];
	} else {
	    throw new FUIException("Superstate index out of bounds: " + i);
	}
    }

    public void addSuperstate(String name) throws FUIException {
	if (!isName(name)) {
	    addSuperstateToMachine(name);
	} else {
	    throw new NamespaceException(name);
	}
    }

    protected void addSuperstateToMachine(String name) {
	superstates = ArrayUtils.addToArray(name, superstates);
	inSuper = ArrayUtils.addColumnToArrayTable(inSuper);
	superTrans = ArrayUtils.addRowToArrayTable(superTrans);
	superId = ArrayUtils.addRowToArrayTable(superId);
    }

    public void deleteSuperstate(int i) {
	if (isSuperstateIndex(i) && superstates.length > 1) {
	    deleteSuperstateFromMachine(i);
	}
    }

    protected void deleteSuperstateFromMachine(int i) {
	superstates = ArrayUtils.deleteFromArray(i, superstates);
	inSuper = ArrayUtils.deleteColumnFromArrayTable(i, inSuper);
	superTrans = ArrayUtils.deleteRowFromArrayTable(i, superTrans);
	superId = ArrayUtils.deleteRowFromArrayTable(i, superId);
    }

    public void renameSuperstate(int i, String name) throws FUIException {
	if (isSuperstateIndex(i) && !name.equals(events[i])) {
	    if (!isName(name)) {
		superstates[i] = name;
	    } else {
		throw new NamespaceException(name);
	    }
	}
    }


    // *************** SET MEMBERSHIP METHODS ***************

    public boolean isElement(int state, int sup) {
	return isStateIndex(state) && isSuperstateIndex(sup) && inSuper[state][sup];
    }

    public boolean[] superstates(int state) {
	return inSuper[state];
    }

    public int[] superstatesIndices(int state) {
	return ArrayUtils.toSet(inSuper[state]);
    }

    public boolean[] elements(int sup) throws FUIException {
	if (isSuperstateIndex(sup)) {
	    return ArrayUtils.column(inSuper, sup);
	} else {
	    throw new FUIException("Superstate index out of bounds: " + sup);
	}
    }

    public int[] elementIndices(int sup) throws FUIException {
	if (isSuperstateIndex(sup)) {
	    return ArrayUtils.trueInColumn(sup, inSuper);
	} else {
	    throw new FUIException("Superstate index out of bounds: " + sup);
	}
    }

    public void setIncluded(int state, int sup, boolean val) {
	if (isStateIndex(state) && isSuperstateIndex(sup)
	    && inSuper[state][sup] != val) {
	    
	    setIncludedInSuperstate(state, sup, val);
	}
    }

    public void toggleIncluded(int state, int sup) {
	if (isStateIndex(state) && isSuperstateIndex(sup)) {
	    setIncludedInSuperstate(state, sup, !inSuper[state][sup]);
	}
    }

    protected void setIncludedInSuperstate(int s, int c, boolean val) {
	inSuper[s][c] = val;
	if (val) {
	    // Add super transitions to state
	    for (int e = 0; e < events.length; e++) {
		if (superTrans[c][e] != null) {
		    for (int i = 0; i < superTrans[c][e].length; i++) {
			if (!isTransition(s, e, superTrans[c][e][i])) {
				trans = ArrayUtils.addArrayTableEntry(s, e, superTrans[c][e][i], trans);
			}
		    }
		}
		
		// Add super identity transition to state
		if (superId[c][e]) {
		    trans = ArrayUtils.addArrayTableEntry(s, e, s, trans);
		}
	    }
	}
    }


    // *************** RANGE METHODS ***************

    public boolean[] superRange(int sup, int event) {
        return ArrayUtils.toPredicate(superTrans[sup][event], superstates.length);
    }

    public int[] superRangeIndices(int sup, int event) {
        return superTrans[sup][event];
    }

    // Deprecate?
    public int[][] superRangeIndices(int sup) {
        return superTrans[sup];
    }
    
    /*
    public int[][][] superRangeIndices() {
	return superTrans;
    }
    */

    // *************** DOMAIN METHODS ***************

    public boolean[] superDomain(int event, int state) {
	
	boolean[] domain = new boolean[superstates.length];
	
	for (int i = 0; i < domain.length; i++) {
	    if (ArrayUtils.inArray(state, superRangeIndices(i, event))) {
		domain[i] = true;
	    }
	}
	return domain;
    }

    public int[] superDomainIndices(int event, int state) {
	return ArrayUtils.toSet(superDomain(event, state));
    }


    // *************** MAP METHODS ***************

    public boolean[] superMaps(int domain, int range) {

	boolean[] maps = new boolean[events.length];

	for (int i = 0; i < maps.length; i++) {
	    if (ArrayUtils.inArray(range, superRangeIndices(domain, i))) {
		maps[i] = true;
	    }
	}
	return maps;
    }

    public int[] superMapsIndices(int domain, int range) {
	return ArrayUtils.toSet(superMaps(domain, range));
    }


    // *************** SUPER TRANSITION METHODS ***************

    public boolean isSuperTransition(int sup, int event, int newState) {
	if (isSuperstateIndex(sup) && isEventIndex(event) && isStateIndex(newState)) {
	    return ArrayUtils.isArrayTableEntry(sup, event, newState, superTrans);
	} else {
	    return false;
	}
    }

    public void setSuperTransition(int sup, int event, int range, boolean val) {

	if (isSuperstateIndex(sup) && isEventIndex(event) && isStateIndex(range)) {
	    if (val) {
		if (!ArrayUtils.isArrayTableEntry(sup, event, range, superTrans)) {
		    setSuperTrans(sup, event, range, true);
		}
	    } else {
		setSuperTrans(sup, event, range, false);
	    }
	}
    }

    public void toggleSuperTransition(int sup, int event, int range) {
	
	if (isSuperstateIndex(sup) && isEventIndex(event) && isStateIndex(range)) {
	    setSuperTrans(sup, event, range, !ArrayUtils.isArrayTableEntry(sup, event, range, superTrans));
	}
    }

    protected void setSuperTrans(int sup, int event, int range, boolean val) {

	if (val) {
	    superTrans = ArrayUtils.addArrayTableEntry(sup, event, range, superTrans);

	    for (int i = 0; i < states.length; i++) {
		if (inSuper[i][sup] && !isTransition(i, event, range)) {
		    trans = ArrayUtils.addArrayTableEntry(i, event, range, trans);
		}
	    }

	} else {
	    superTrans = ArrayUtils.deleteArrayTableEntry(sup, event, range, superTrans);
	}
    }


    /*
    public void deleteSuperTransitions(int sup, int event) {
	if (isSuperstateIndex(sup) && isEventIndex(event)) {
	    superTrans[sup][event] = null;
	}
    } 
    
    public void deleteSuperTransitions(int sup) {
	if (isSuperstateIndex(sup)) {
	    superTrans[sup] = new int[events.length][];
	}
    } 
    */


    // *************** IDENTITY TRANSITION METHODS ***************

    public boolean isSuperIdentity(int sup, int evt) {
	return isSuperstateIndex(sup) && isEventIndex(evt) && superId[sup][evt];
    } 

    public boolean[] superIdentities(int sup) {
	return superId[sup];
    } 

    public void setSuperIdentity(int sup, int evt, boolean val) {
	if (isSuperstateIndex(sup) && isEventIndex(evt)) {
	    superId[sup][evt] = val;
	}	
    }


    // *************** OVERRIDING FSM METHODS ***************

    protected void addStateToMachine(String name) {
	super.addStateToMachine(name);
	inSuper = ArrayUtils.addRowToArrayTable(inSuper);
    }

    protected void deleteStateFromMachine(int i) {
	inSuper = ArrayUtils.deleteRowFromArrayTable(i, inSuper);
	superTrans = ArrayUtils.deleteAllFromArrayTable(i, superTrans);
	super.deleteStateFromMachine(i);	
    }

    protected void addEventToMachine(String name) {
	super.addEventToMachine(name);
	superTrans = ArrayUtils.addColumnToArrayTable(superTrans);
	superId = ArrayUtils.addColumnToArrayTable(superId);
    }

    protected void deleteEventFromMachine(int i) {
	superTrans = ArrayUtils.deleteColumnFromArrayTable(i, superTrans);
	superId = ArrayUtils.deleteColumnFromArrayTable(i, superId);
	super.deleteEventFromMachine(i);
    }

    /* Don't remove state transitions inherited from a superstate.
     */
    protected void deleteTransFromMachine(int domain, int event, int range) {
	    
	boolean inherited = false;
	int[] cs = superstatesIndices(domain);
	if (cs != null) {
	    for (int i = 0; i < cs.length; i++) {
		if (isSuperTransition(cs[i], event, range)
		    || (domain == range && isSuperIdentity(cs[i], event))) {
		    
		    inherited = true;
		    break;
		}
	    }
	}
	
	if (!inherited) {
	    super.deleteTransFromMachine(domain, event, range);
	}
    }

}
