package org.uclic.matrix;

import org.jdom.Element;

/**
 * An integer matrix with some basic operations.
 *
 * @author Jeremy Gow
 * @version 0.3.2
 */
public class Matrix {

    protected int[][] entries;
    protected int     rows;
    protected int     cols;

    public final static String MATRIX = "matrix";
    public final static String ENTRY  = "entry";
    public final static String ROW    = "row";

    public Matrix(int r, int c) {
        entries = new int[r][c];
        rows = r;
        cols = c;
    }

    public Matrix(int s) {
        this(s, s);
    }

    public Matrix(int[][] e) {
        entries = e;
        rows = e.length;
        cols = e[0].length;
    }

    /*
     * Get matrix entry.
     */
    public int getEntry(int r, int c) {
        if (isEntry(r, c)) {
            return entries[r][c];
        } else {
            return 0;
        }
    }

    /*
     * Set a matrix entry.
     */
    public void setEntry(int r, int c, int e) {
        if (isEntry(r, c)) {
            entries[r][c] = e;
        }
    }

    public int rows() {
        return rows;
    }

    public int cols() {
        return cols;
    }

    public boolean isEntry(int r, int c) {
        return (-1 < r && r < rows() && -1 < c && c < cols());
    }

    public String dimensions() {
	return rows() + "x" + cols();
    }

    public void setRow(int i, int[] r) {
        if (!(i < 0) && i < rows() && r.length == cols()) {
            entries[i] = r;
        }
    }

    public boolean equals(Matrix other) {
	boolean equal = (rows() == other.rows() && cols() == other.cols());
	if (equal) {
	    for (int i = 0; i < rows(); i++) {
		for (int j = 0; j < cols(); j++) {
		    if (getEntry(i, j) != other.getEntry(i, j)) {
			equal = false;
			break;
		    }
		}
		if (!equal) {
		    break;
		}
	    }
	}
	return equal;
    }

    public boolean[] equalRows(Matrix other) {

	boolean[] eval = new boolean[rows()];
	
	if (isSameSize(other)) {
	    for (int r = 0; r < rows(); r++) {
		eval[r] = true;

		for (int c = 0; c < cols(); c++) {
		    if (getEntry(r, c) != other.getEntry(r, c)) {
			eval[r] = false;
			break;
		    }
		}
	    } 
	}

	return eval;
    }

    public boolean[] equalColumns(Matrix other) {

	boolean[] eval = new boolean[cols()];
	
	if (isSameSize(other)) {
	    for (int c = 0; c < cols(); c++) {
		eval[c] = true;

		for (int r = 0; r < rows(); r++) {
		    if (getEntry(r, c) != other.getEntry(r, c)) {
			eval[c] = false;
			break;
		    }
		}
	    }
	}

	return eval;
    }


    /**
     * Transpose
     */
    public Matrix transpose() {
        Matrix trans = new Matrix(cols(), rows());

        for (int i = 0; i < rows(); i++) {
            for (int j = 0; j < cols(); j++) {
                trans.setEntry(j, i, getEntry(i, j));
            }
        }
        return trans;
    }

    /**
     * Matrix Comparison
     */
    public boolean isSameSize(Matrix other) {
        return (rows() == other.rows() && cols() == other.cols());
    }

    public boolean isComposable(Matrix other) {
        return (cols() == other.rows());
    }

    public boolean isSquare(int k) {
        return (rows() == k && cols() == k);
    }

    public boolean isIdentity() {
	boolean id = (rows() == cols());

	if (id) {
	    for (int i = 0; i < rows(); i++) {
		for (int j = 0; j < cols(); j++) {
		    if (getEntry(i, j) != 0) {
			if ((i != j) || (getEntry(i, j) != 1)) {
			    return false;
			}
		    }
		}
	    }
	}
	
	return id;
    }

    /**
     * Matrix Addition
     */
    public Matrix add(Matrix other) throws WrongSize {

        if (isSameSize(other)) {
            Matrix sum = new Matrix(rows(), cols());
            int entry;

            // Compute each i,j entry
	    for (int i = 0; i < rows(); i++) {
		for (int j = 0; j < cols(); j++) {
		    // Add entries
		    entry = getEntry(i, j) + other.getEntry(i, j);
		    sum.setEntry(i, j, entry);
		}
	    }
	    return sum;
        } else {
            throw new WrongSize("Matrices cannot be added.");
        }
    }

    /**
     * Matrix Multiplication
     */
    public Matrix multiply(Matrix matrix) throws WrongSize {
	Matrix product = new Matrix(rows(), matrix.cols());
	multiply(matrix, product);
	return product;
    }

    protected void multiply(Matrix multiplier, Matrix result) throws WrongSize {

	if (isComposable(multiplier)) {

            int entry;

	    // Compute each i,j entry
	    for (int i = 0; i < rows(); i++) {
		for (int j = 0; j < multiplier.cols(); j++) {
		    // Row * Column
		    entry = 0;
		    for (int k = 0; k < cols(); k++) {
                            entry += getEntry(i, k) * multiplier.getEntry(k, j);
		    }
		    result.setEntry(i, j, entry);
		}
	    }
	} else {
            throw new WrongSize("Matrices cannot be multiplied: " + dimensions() + " * " + multiplier.dimensions());
        }
    }

    public Matrix tryMultiply(Matrix matrix) {
	try {
	    return multiply(matrix);
	} catch (WrongSize ws) {
	    return this;
	}
    }

    public Matrix multiply(int constant) {
	Matrix multiple = new Matrix(rows(), cols());
	multiply(constant, multiple);
	return multiple;
    }

    public Matrix multiply(int constant, Matrix result) {

	for (int i = 0; i < rows(); i++) {
	    for (int j = 0; j < cols(); j++) {
		result.setEntry(i, j, getEntry(i, j) * constant);
	    }
	}
	return result;
    }

    
    public boolean singular() {
	return (determinant() == 0);
    }

    public int determinant() {

	int sum = 0;
	
	for (int r = 0; r < rows(); r++) {
	    sum = sum + cofactorProduct(r);
	}

	return sum;
    }

    public int cofactorProduct(int j) {
	
	Minor minor = new Minor(this, 1, j);
	return minor.determinant() * (-1)^j * getEntry(1, j);
    }

    class Minor extends Matrix {
	
	private Matrix  major;
	private int     elimRow;
	private int     elimCol;

	public Minor(Matrix m, int r, int c) {
	    super(m.rows() - 1, m.cols() - 1);
	    entries = null;
	    major = m;
	    elimRow = r;
	    elimCol = c;
	}

	private int adjust(int i, int e) {
	    if (i >= e) {
		i--;
	    }
	    return i;
	}

	public int getEntry(int i, int j) {
	    i = adjust(i, elimRow);
	    j = adjust(j, elimCol);
	    return major.getEntry(i, j);
	}

	public void setEntry(int i, int j, int v) {
	    i = adjust(i, elimRow);
	    j = adjust(j, elimCol);
	    major.setEntry(i, j, v);
	}
    }


    // To implement
    public Matrix stack(int i) {
	return new Matrix(1);
    }

    // To implement
    public Matrix append(int i) {
	return new Matrix(1);
    }

    // To implement
    public Matrix stack(Matrix m) {
	return new Matrix(1);
    }

    // To implement
    public Matrix append(Matrix m) {
	return new Matrix(1);
    }



    /*
     * Write as XML
     */
    public Element toXML() {

	Element entry, row, matrix = new Element(MATRIX);

	for (int i = 0; i < rows(); i++) {
	    
	    row = new Element(ROW);
	    matrix.addContent(row);

	    for (int j = 0; j < rows(); j++) {
	   
		entry = new Element(ENTRY);
		row.addContent(entry);
		entry.addContent(Integer.toString(getEntry(i, j)));
	    } 
	}

	return matrix;
    } 
  
}
