package org.uclic.maui;

import java.awt.event.*;
import java.io.*;
import java.text.DateFormat;
import java.util.Date;
import javax.swing.*;
import javax.swing.filechooser.FileFilter;
import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;
import org.uclic.fui.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;

/**
 * Provides various file operations for MAUI. 
 *
 * @author Jeremy Gow
 * @version 0.4
 */
public class FileOperator extends JFileChooser implements ActionListener {

    private MAUI app;
    private int  fChoosen;

    public final static String DTD_PREFIX  = "file://";
    public final static String OPEN_BUTTON = "Open";
    public final static String OPEN_DESIGN = "Open Design File";
    public final static String SAVE_BUTTON = "Save";
    public final static String SAVE_DESIGN = "Save Design File";
    public final static String SAVE_LOG    = "Save Log File";

    public FileOperator(MAUI a) {
        super();
        app = a;
	addChoosableFileFilter(new SuffixFilter(".fui", "FUI Files"));
        addChoosableFileFilter(new SuffixFilter(".xml", "XML Files"));
        addChoosableFileFilter(new SuffixFilter(".log", "Log Files"));
    }

    public void chooseFile() {
        fChoosen = showOpenDialog(app); 
    }
	
    public void modelFilter() {
        setFilter(1);
    }

    public void logFilter() {
        setFilter(3);
    }

    public void setFilter(int i) {
        setFileFilter(getChoosableFileFilters()[i]);
    }
	
    /**
     * User Operations
     */
    public void openModel() {
        setApproveButtonText(OPEN_BUTTON);
        setDialogTitle(OPEN_DESIGN);
        modelFilter();
        chooseFile();
        if (fChoosen == APPROVE_OPTION) {
	    try {
		app.setDesign(readModel());
	    } catch (FileNotFoundException e1) {
		app.error(error("reading") + "File not found.");
	    } catch (JDOMException e2) {
		app.error(error("reading") + e2.getMessage() + "\n\nCheck that the DTD reference points to the local DTD or to\nhttp://www.uclic.ucl.ac.uk/usr/jgow/maui/fui.dtd");
	    } catch (FUIException e3) {
		app.error(error("reading") + e3.getMessage());
		e3.printStackTrace();
	    } catch (Exception e4) {
		app.error("Unknown error: Unable to read model file.");
		e4.printStackTrace();
	    } 
	} 
    }

    public void saveModel() {
        setApproveButtonText(SAVE_BUTTON);
        setDialogTitle(SAVE_DESIGN);
        modelFilter();
        chooseFile();
        if (fChoosen == APPROVE_OPTION) {
            writeModel(app.design());
        }
    }
  
    public void saveLog(String log) {
        setApproveButtonText(SAVE_BUTTON);
        setDialogTitle(SAVE_LOG);
        logFilter();    
        chooseFile();
        if (fChoosen == APPROVE_OPTION) {      
            writeLog(log);
        }
    }

    private void rememberFile(File file) {
	app.setPref(MAUI.CURRENT_PROP, file.getAbsolutePath());    
    }

    /**
     *  File Operations
     */
    public FUIDesign readModel() throws FileNotFoundException, JDOMException, FUIException, IOException {

	File file = getSelectedFile();
	SAXBuilder builder = new SAXBuilder(true);
	//builder.setXMLFilter(new MAUIDTDFilter());
        Document xml = builder.build(new FileInputStream(file));
	String systemDTD, shortDTD;

	systemDTD = xml.getDocType().getSystemID();
        int i = systemDTD.lastIndexOf(System.getProperty("file.separator"));

        if (i == -1) {
            shortDTD = systemDTD;
        } else {
            shortDTD = systemDTD.substring(++i);
        }

	FUIReader rdr = new FUIReader(xml.getRootElement());
	FUIDesign dgn = rdr.read();
	rememberFile(file);

	return dgn;
    }

    /* Tried to filter DTD, so that we could ensure local fui.dtd is used.
       But for some reason notationDecl is never called.
     
    private class MAUIDTDFilter extends XMLFilterImpl {
	public void notationDecl(String name, String publicId, String systemId) throws SAXException {
	    super.notationDecl(name, publicId, systemId);
	    System.out.println("notationDecl");
	}

	public void startDocument() throws SAXException {
	    super.startDocument();
	    System.out.println("startDocument");
       	}
    }
    */

    public void writeModel(FUIDesign design) {
        try {    	
	    //String dir = app.home() + System.getProperty("file.separator");
	    FUIWriter wtr = new FUIWriter(design);
	    Element elem = wtr.write();
	    String sep = System.getProperty("file.separator");
	    String dtd = DTD_PREFIX + app.home() + sep + MAUI.MODEL_DIR + sep + MAUI.DTD;
	    Document doc = new Document(elem, new DocType("fui", dtd));
	    
	    XMLOutputter xo = new XMLOutputter();
	    xo.setFormat(Format.getPrettyFormat());
	    File file = getSelectedFile();
	    FileOutputStream o = new FileOutputStream(file);
	    xo.output(doc, o);
            o.close();
	    rememberFile(file);

        } catch (FileNotFoundException e) {
            app.error(error("writing") + "No such file.\n");
        } catch (IOException e) {
            app.error(error("writing") + e.getMessage() + "\n");
        }
    }

    public void writeLog(String log) {
        try {
	   
            FileWriter writer = new FileWriter(getSelectedFile());
            Date etad = new Date(System.currentTimeMillis());
            int s = DateFormat.SHORT;
            String time = DateFormat.getTimeInstance(s).format(etad);
            String date = DateFormat.getDateInstance(s).format(etad);
 			
            writer.write("Model: " + app.design().name() + "\n");
            writer.write("Logged: " + time + " " + date + "\n\n");
            writer.write(log);
            writer.close();
					
        } catch (IOException e) {
            app.error(error("writing") + e.getMessage() + "\n");
        }
    }
	
    private String error(String verb) {
        String er = "Error " + verb + " ";
        String ror = getSelectedFile().getAbsolutePath();

        return er + ror + "\n";
    }
	
    /**
     *  Actions
     */
    public void actionPerformed(ActionEvent e) {

        String com = e.getActionCommand();

        if (com.equals(MAUIMenu.OPEN_FILE)) {
            openModel();
        } else if (com.equals(MAUIMenu.SAVE_FILE)) {
            saveModel();
        } 
    }

    /**
     * File Filters
     */  
    class SuffixFilter extends javax.swing.filechooser.FileFilter {	
  	
        private String fDot;
        private String fBlurb;
  	
        public SuffixFilter(String pDot, String pBlurb) {
            super();
            fDot = pDot;
            fBlurb = pBlurb;
        }
  		
        public boolean accept(File pFile) {
            String name = pFile.getName();
            boolean okFile = name.endsWith(fDot);
            boolean isDir = pFile.isDirectory();

            return (okFile || isDir);
        }

        public String getDescription() {
            return fBlurb;
        }
    }
}

