package org.uclic.maui.analyse;

import java.awt.*;
import java.awt.event.*;
import java.beans.PropertyVetoException;
import java.util.Date;
import java.util.Observer;
import java.util.Observable;
import java.util.Vector;
import javax.swing.*;
import javax.swing.table.*;
import org.jdom.Element;
import org.uclic.matrix.Matrix;
import org.uclic.matrix.WrongSize;
import org.uclic.algebra.*;
import org.uclic.algebra.property.*;
import org.uclic.fui.*;
import org.uclic.maui.*;

/**
 * @author Jeremy Gow
 * @version 0.4
 */
public class PropertyTable extends JTable {

    private MAUI               maui;
    private PropertyTableModel model;
    private DialogCellEditor   editor;    

    private ImageIcon     tick;
    private ImageIcon     cross;
    private ImageIcon     unknown;
    private final static String TICK_FILE    = "tick.gif";
    private final static String CROSS_FILE   = "cross.gif";
    private final static String UNKNOWN_FILE = "unknown.jpg";

    private final static String[] columnNames = {"Property", "Status", "States", "Lock"};
    private final static int PROPERTY_COLUMN = 0;
    private final static int STATUS_COLUMN   = 1;
    private final static int PARTIAL_COLUMN  = 2;
    private final static int LOCK_COLUMN     = 3;

    private final static int PROPERTY_WIDTH  = 350;
    private final static int ROW_HEIGHT      = 30;
    
    private final static String PROPERTY_TIP = "Click twice to edit";
    private final static String PARTIAL_TIP = "Click twice for details";
    
    public PropertyTable(MAUI m) {
	super();
	maui = m;
	model = new PropertyTableModel();
	setModel(model);
	setCellSelectionEnabled(true);
	setRowHeight(ROW_HEIGHT);

	TableColumn prop = getColumnModel().getColumn(PROPERTY_COLUMN);
	TableColumn part = getColumnModel().getColumn(PARTIAL_COLUMN);
	prop.setPreferredWidth(PROPERTY_WIDTH);	
	prop.setCellRenderer(new PropertyRenderer());
	part.setCellRenderer(new PropertyRenderer());

        String sep = System.getProperty("file.separator");
	String path = maui.home() + sep + MAUI.ICONS + sep;
        tick = new ImageIcon(path + TICK_FILE);
        cross = new ImageIcon(path + CROSS_FILE);
        unknown = new ImageIcon(path + UNKNOWN_FILE);

    }


    private class PropertyRenderer extends JLabel implements TableCellRenderer {

	public PropertyRenderer() {super();}

	public Component getTableCellRendererComponent(JTable t, Object v, boolean s, boolean f, int r, int c) { 
	    setText(v.toString());
	    setHorizontalAlignment(JLabel.CENTER);
	    setFont(t.getFont());
	    
	    if (c == PROPERTY_COLUMN) {
		setToolTipText(PROPERTY_TIP);
	    
	    } else if (c == PARTIAL_COLUMN) {
		setToolTipText(PARTIAL_TIP);
	    }

	    return this;
	}
    }

    /**
     * Notify table that design has changed.
     */
    public void designChanged(Vector edits) {
	if (edits.contains(new Integer(MAUI.MODEL_EDIT))
	    || edits.contains(new Integer(MAUI.RENAME_EDIT))) {
	    model.fireTableDataChanged();
	}
    }



    /**
     * Return a editor for a cell.
     * Returns specialised editors for PROPERTY_COLUMN and PARTIAL_COLUMN.
     */
    public TableCellEditor getCellEditor(int row, int col) {

	switch (col) {
	    
	case PROPERTY_COLUMN: 
	    // Return property editor.
	    return new DialogCellEditor(new PropertyEditor(maui));

	case PARTIAL_COLUMN:
	    // Return partial data dialog.
	    return new DialogCellEditor(new PartialData(maui));
	   
	default:
	    // Return default editor for column.
	    return super.getCellEditor(row, col);
	}
    }

    private class DialogCellEditor extends AbstractCellEditor implements TableCellEditor, ActionListener {

	private CellDialog dialog;
	private JButton    cell;
	private boolean    once;

	private final static String EDIT = "edit";
    
	public DialogCellEditor(CellDialog d) {
	    super();
	    dialog = d;
	    d.registerEditor(this);
	    once = false;
	    editor = this;
	    
	    // Create cell button
	    cell = new JButton();
	    cell.setActionCommand(EDIT);
	    cell.addActionListener(this);
	    cell.setBorderPainted(false);
	    
	    cell.setFont(getFont());
	    cell.setBackground(getSelectionBackground());
	    cell.setToolTipText(dialog.getTip());
	}

	public void actionPerformed(ActionEvent e) {

	    String cmd = e.getActionCommand();
       
	    if (once && cmd.equals(EDIT)) {
		maui.setEnabled(false);
		dialog.setVisible(true);
	    }

	    once = true;
	}

	public Component getTableCellEditorComponent(JTable t, Object v, boolean s, int r, int c) {
	    dialog.configure(r);
	    cell.setText(model.getValueAt(r, c).toString());
	    return cell;
	}

	public Object getCellEditorValue() {
	    return dialog.getValue();
	}

	public boolean stopCellEditing() {
	    dialog.setVisible(false);
	    maui.setEnabled(true);	    
	    clearSelection();
	    return super.stopCellEditing();
	}

	public void cancelCellEditing() {
	    dialog.setVisible(false);
	    maui.setEnabled(true);	    
	    clearSelection();
	    super.cancelCellEditing();
	}
    }

    /*
     * Add a new blank property to the table.
     */
    public void newProperty() {

	int newRow;

	Specification spec = maui.design().spec();

	if (spec.isEmpty()) {
	    newRow = 0;		
	    spec.addProperty();
		
	} else if (spec.isEmpty() || getSelectedRowCount() == 0) {
	    newRow = spec.size();
	    spec.addProperty();
				
	} else {
	    int[] rows = getSelectedRows();
	    newRow = rows[rows.length - 1] + 1;
	    spec.addProperty(newRow);
	}

	model.fireTableRowsInserted(newRow, newRow);
	maui.designChanged(MAUI.SPEC_EDIT);
    }

    /*
     * Delete one or more properties from the table.
     */
    public void deleteProperties() {

	String message, suffix;
	int rows[] = getSelectedRows();
	int s = rows.length;

	if (s > 0) {
	    if (s == 1) {
		suffix = "this property?";
	    } else {
		suffix = "these properties?";
	    }
	    message = "Are you sure you want to delete " + suffix;

	    int o = JOptionPane.showConfirmDialog(maui, message, "Are you SURE?",
						  JOptionPane.YES_NO_OPTION,
						  JOptionPane.QUESTION_MESSAGE);
		
	    if (o == JOptionPane.YES_OPTION) {
		for (int i = 0; i < rows.length; i++) {
		    int r = rows[i] - i;
		    maui.design().spec().removeProperty(r);
		    model.fireTableRowsDeleted(r, r);
		}
		editor.cancelCellEditing();
		maui.designChanged(MAUI.SPEC_EDIT);
	    }
	}
    }

    public void evaluate() {
	maui.design().spec().evaluate();
	for (int i = 0; i < maui.design().spec().size(); i++) {
	    model.fireTableCellUpdated(i, 1);
	    model.fireTableCellUpdated(i, 2);
	}
    }

    private void lock() {
	    
	int[] props = getSelectedRows();

	for (int i = 0; i < props.length; i++) {
	    maui.design().spec().setLock(props[i], !maui.design().spec().isLocked(props[i]));
	    model.fireTableCellUpdated(props[i], 3);
	}
    }
    
    private class PropertyTableModel extends AbstractTableModel {
		
	public PropertyTableModel() {
	    super();
	}
	
	public int getColumnCount() {
	    return columnNames.length;
	}
	
	public int getRowCount() {
	    return maui.design().spec().size();
	}
	
	public Class getColumnClass(int c) {
	    return getValueAt(0, c).getClass();
	}
	
	public String getColumnName(int col) {
	    return columnNames[col];
	}
	
	public boolean isCellEditable(int row, int col) {
	    return (col != STATUS_COLUMN);
	}

	public Object getValueAt(int row, int col) {

	    Specification spec = maui.design().spec();

	    Object value = null;

	    switch (col) {

	    case PROPERTY_COLUMN: 
		value = spec.property(row).termString();
		break;

	    case STATUS_COLUMN:
		if (spec.valueKnown(row)) {
		    if (spec.value(row)) {
			value = tick;
		    } else {
			value = cross;
		    }
		} else {
		    value = unknown;
		}
		break;
		
	    case PARTIAL_COLUMN:
		if (spec.percentKnown(row)) {
		    value = spec.percent(row) + "%";
		} else {
		    value = "-";
		}
		break;
		
	    case LOCK_COLUMN:
		value = spec.isLockedObject(row);
		break;
	    }

	    return value;
	}

	public void setValueAt(Object val, int row, int col) {

	    switch (col) {

	    case PROPERTY_COLUMN:
		PropertyTerm p = (PropertyTerm) val;
		maui.design().spec().setProperty(p, row);
		maui.designChanged(MAUI.SPEC_EDIT);
		break;
		
	    case LOCK_COLUMN:
		Boolean b = (Boolean) val;
		maui.design().spec().setLock(row, b.booleanValue());
		break;
	    }
	}
    }
}
