package org.uclic.maui.analyse;

import java.awt.*;
import java.awt.event.*;
import java.util.Enumeration;
import java.util.EventObject;
import java.util.Iterator;
import java.util.Vector;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;
import org.uclic.algebra.*;


/**
 * Editor for event algebra properties.
 *
 * @author Jeremy Gow
 * @version 0.3
 */
public class TermTree extends JTree {

    private DefaultTreeModel model;
    
    public TermTree(MutableTreeNode node, CellEditorListener listener) {
	super();
	getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
	setEditable(true);
	setShowsRootHandles(true);

	// Tree Model
	model = new DefaultTreeModel(node);
	setModel(model);

	// Cell Renderer
	DefaultTreeCellRenderer render = new DefaultTreeCellRenderer();
	render.setOpenIcon(null);
	render.setClosedIcon(null);
	render.setLeafIcon(null);
	setCellRenderer(render);

	// Cell Editor
	TermEditor editor = new TermEditor(this); 
	editor.addCellEditorListener(listener);
       	setCellEditor(editor);

	expandAll();
    }


    /*
     * Expand all nodes of the term tree.
     */
    public void expandAll() {
        TreeNode root = (TreeNode) model.getRoot();
	expandAll(new TreePath(root));
    }

    /*
     * Expand all nodes below and including the given path.
     */
    private void expandAll(TreePath parent) {

        TreeNode node = (TreeNode) parent.getLastPathComponent();
        if (node.getChildCount() >= 0) {
	    Enumeration en = node.children();
            while (en.hasMoreElements()) {
                TreeNode n = (TreeNode) en.nextElement();
                TreePath path = parent.pathByAddingChild(n);
                expandAll(path);
            }
        }
	expandPath(parent);
    }

    /*
     * Replace the given node of the term tree.
     */
    public void replaceSubterm(Term oldNode, Term newNode) {

	Term parent = (Term) oldNode.getParent();
		
	if (parent != null) {
	    // Replace child node
	    int index = parent.getIndex(oldNode);
	    model.removeNodeFromParent(oldNode);
	    model.insertNodeInto(newNode, parent, index);
	    expandAll();
	   
	} else {
	    // Replace root node
	    try {
		model.setRoot(newNode);
	    } catch (NullPointerException e) {}
	}
	repaint();	
    }   

    class TermEditor implements TreeCellEditor, ActionListener {

	private TermTree   tree;
	private Term       node;
	private String[]   terms;
	private JComboBox  box;
	private JTextField text;
	private Vector     listeners;
	
	public TermEditor(TermTree t) {
	    tree = t;
	    listeners = new Vector();
	    buildBox(new String[0]);
	    text = new JTextField();
	    box.setBackground(Color.WHITE);
	    text.addActionListener(this);
	}

	public void buildBox(String[] items) {
	    box = new JComboBox(items);
	    box.setBackground(Color.WHITE);
	    box.addActionListener(this);
	}

	public void actionPerformed(ActionEvent e) {

	    int i = 0;

	    if (node.termType() == Term.TEXT_TYPE) {
		i = Integer.parseInt(text.getText()); 
	    } else {
		i = box.getSelectedIndex();		
	    }
	    tree.replaceSubterm(node, node.buildTerm(i));
	    stopCellEditing();
	}

	public Object getCellEditorValue() {
	    return box.getSelectedItem();
	}
	
	public Component getTreeCellEditorComponent(JTree tree, Object value, boolean isSelected,
						    boolean expanded, boolean leaf, int row) {

	    node = (Term) value;		
	    
	    if (node.termType() == Term.TEXT_TYPE) {
		text.setText(node.toString());
		return text;
	    } else {
		buildBox(node.terms());
		return box;
	    }
	}

	public boolean isCellEditable(EventObject anEvent) {
	    return true;
	}

	public boolean shouldSelectCell(EventObject anEvent) {
	    return true;
	}

	/*
	 * Cell editor listeners.
	 */

	public void addCellEditorListener(CellEditorListener l) {
	    listeners.add(l);
	}

	public void removeCellEditorListener(CellEditorListener l) {
	    listeners.remove(l);
	}

	public boolean stopCellEditing() {

	    Iterator itr = listeners.iterator();
	    CellEditorListener cel;

	    while (itr.hasNext()) {
		cel = (CellEditorListener) itr.next();
		cel.editingStopped(new ChangeEvent(this));
	    }
	    return true;
	}

	public void cancelCellEditing() {

	    Iterator itr = listeners.iterator();
	    CellEditorListener cel;

	    while (itr.hasNext()) {
		cel = (CellEditorListener) itr.next();
		cel.editingCanceled(new ChangeEvent(this));
	    }
	}
    }
}
