package org.uclic.maui.edit;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import org.uclic.fui.*;
import org.uclic.maui.*;
import org.uclic.util.ArrayUtils;

/**
 * A graphical editor for the formal user interface model.
 *
 * @author Jeremy Gow
 * @version 0.4
 *
 * Setting initial state?
 */
public class Editor extends MAUITab implements ActionListener {

    // Lists
    public SuperPanel superPanel;
    public StatePanel statePanel;
    public TransPanel transPanel;
    public EventPanel eventPanel;
    public RangePanel rangePanel;
    public SignsPanel signsPanel;

    private JLabel     label;
    private JTextField name;
    private ListPanel  last;

    private boolean listening;

    public final static String SUPERS = "Superstates";
    public final static String STATES = "States";
    public final static String TRANS  = "Transitions";
    public final static String EVENTS = "Events";
    public final static String RANGE  = "To States";
    public final static String SIGNS  = "Signs";

    public final static String RENAME      = "Rename";
    public final static String SUPER_LABEL = "Superstate: ";
    public final static String STATE_LABEL = "State: ";
    public final static String EVENT_LABEL = "Event: ";
    public final static String RANGE_LABEL = "State: ";
    public final static String SIGN_LABEL  = "Sign: ";

    private final static String TITLE = "Editor";
    private final static String ICON  = "Edit24.gif";
    private final static String TIP   = "Edit the interface model";

    public Editor(MAUI a) {
	super(TITLE, a, ICON, TIP);
	setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
	buildTab();
    }

    private void buildTab() {

	listening = false;

	JPanel lists = new JPanel();
	lists.setLayout(new BoxLayout(lists, BoxLayout.X_AXIS));
	superPanel = new SuperPanel(lists, this);
	statePanel = new StatePanel(lists, this);
	transPanel = new TransPanel(lists, this);
	signsPanel = new SignsPanel(lists, this);

	label = new JLabel();
	name = new JTextField(20);
	name.addActionListener(this);
	name.setActionCommand(RENAME);
	JButton rename = new JButton(RENAME);
	rename.addActionListener(this);
	rename.setActionCommand(RENAME);
	
	JPanel bar = new JPanel();
	bar.add(label);
	bar.add(name);
	bar.add(rename);
	bar.setMaximumSize(new Dimension(MAUI.XDIM, 20));

	add(lists);
	add(bar);

	superPanel.populateList();
	statePanel.populateList(); // Automatically populates rangePanel
	eventPanel.populateList();
	signsPanel.populateList();

	//statePanel.deselect();
	selectPanel(statePanel);
	listening = true;
    }

    public void rebuildTab() {
	//System.out.println("Rebuild editor");
	removeAll();
	buildTab();
    }

    public void protectLocks(HighlightList list) {

	algebra().redefined();
	int[] broken = spec().checkLocks();

	if (broken == null) {
	    spec().allUnknown();
	    broadcastChange(MAUI.MODEL_EDIT);

       	} else {	    
	    String prop = "property";
	    StringBuffer props = new StringBuffer();
	    if (broken.length > 1) {
		prop = "properties";
	    }
	    props.append(": " + (spec().property(broken[0]).termString()));
	    for (int i = 1; i < broken.length; i++) {
		props.append(",  " + spec().property(broken[i]).termString());
	    }
	    maui.setStatus("Edit prevented by lock on " + prop + props);
	    design().restoreModel();
	    list.restoreHighlights();
	}
    }

    private class TransPanel extends EditorPanel {
	
	public TransPanel(JPanel p, Editor ed) {
	    super(TRANS, p, ed);	 
	    
	    JPanel t2 = new JPanel();
	    add(t2);
	    eventPanel = new EventPanel(t2, ed);
	    rangePanel = new RangePanel(t2, ed);
	}
    }

    public boolean isTransSelected() {
	return transPanel.isSelected();
    }

    public boolean isListening() {
	return listening;
    }

    public void setListening(boolean l) {
	listening = l;
    }
    

    /* Respond to panel selection.
     */
    public void actionPerformed(ActionEvent e) {
	
	if (listening) {
	    maui.clearStatus();

	    String cmd = e.getActionCommand();
	    
	    if (cmd.equals(SUPERS)) {
		selectPanel(superPanel);
		enablePanels(true, true);

	    } else if (cmd.equals(STATES)) {
		selectPanel(statePanel);
		enablePanels(true, true);

	    } else if (cmd.equals(TRANS)) {
		enablePanels(true, false);
		selectPanel(transPanel);
		signsPanel.list.clearAll();
		
	    } else if (cmd.equals(EVENTS)) {
		eventPanel.select();
		if (!transPanel.isSelected()) {
		    rangePanel.deselect();
		}

	    } else if (cmd.equals(RANGE)) {
		rangePanel.select();
		if (!transPanel.isSelected()) {
		    eventPanel.deselect();
		}

	    } else if (cmd.equals(SIGNS)) {
		enablePanels(false, true);
		selectPanel(signsPanel);
		eventPanel.list.clearAll();
		rangePanel.list.clearAll();
	    
	    } else if (cmd.equals(RENAME)) {
		last.renameItem(name.getText());
	    }
	}
    }

    private void enablePanels(boolean trans, boolean signs) {
	eventPanel.setTitleEnabled(trans);
	rangePanel.setTitleEnabled(trans);
	if (signsPanel.isEnabled() != signs) {
	    signsPanel.setEnabled(signs);
	}
    }

    private void selectPanel(EditorPanel select) {
	selectPanel(select, superPanel);
	selectPanel(select, statePanel);
	selectPanel(select, transPanel);
	selectPanel(select, signsPanel);
	

	if (transPanel.isSelected()) {
	    eventPanel.select();
	    rangePanel.select();    

	} else if (eventPanel.isSelected() && rangePanel.isSelected()) {
	    rangePanel.deselect();
	}
    }

    private void selectPanel(EditorPanel select, EditorPanel panel) {
	if (panel.equals(select)) {
	    panel.select();
	} else {
	    panel.deselect();
	}	
    }

    public void setLastList(ListPanel list) {
	last = list;
	label.setText(last.getItemType());
	name.setText(last.getSelectedItem());
    }

}
