package org.uclic.maui.sim;

import java.awt.*;
import java.awt.event.*;
import java.util.Observer;
import java.util.Observable;
import java.util.Random;
import java.beans.*;
import javax.swing.*;
import javax.swing.border.*;
import java.util.Enumeration;
import org.uclic.fui.*;
import org.uclic.maui.*;

/**
 * A simulation of the interface model.  This consists of the current
 * state (via a changable dropdown menu) and device's buttons.
 * Pressing the buttons changes the state, as does selecting a new
 * state from the menu.
 *
 * @author Jeremy Gow
 * @version 0.4
 */
public class Simulation extends MAUITab implements ActionListener {

    private JComboBox     stateMenu;
    private boolean       listening;
    private int[][]       next;
    private Random        random;
    private JPanel        eventPanel;
    private EventButton[] eventButtons;
    private JList         signs;
    private JTextArea     log;

    public final static String RESET = "Reset";
    public final static String CLEAR = "Clear";
    public final static String SAVE = "Save Log";
    public final static int LOG_WIDTH = 20;
    public final static int LOG_HEIGHT = 36;

    private final static String TITLE = "Simulation";
    private final static String ICON  = "Play24.gif";
    private final static String TIP   = "Simulate the interface";

    public Simulation(MAUI a) {
        super(TITLE, a, ICON, TIP);

	listening = true;
	random = new Random();	
	build();
    }

    private void build() {

	// Build sign panel
	signs = new JList();
	signs.setSelectionMode(ListSelectionModel.SINGLE_INTERVAL_SELECTION);
	signs.setLayoutOrientation(JList.HORIZONTAL_WRAP);
	//signs.setVisibleRowCount(-1);
	
	JScrollPane signScroll = new JScrollPane(signs);
	signScroll.setPreferredSize(new Dimension(500, 100));
	signScroll.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED);
	JPanel signPanel = new JPanel();
	border(signPanel, "Signs");
	signPanel.add(signScroll);
	
	// Build log
	log = new JTextArea(LOG_HEIGHT, LOG_WIDTH);
        JScrollPane logScroll = new JScrollPane(log);
        logScroll.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);

	// Build events/states
	buildEventPanel();
	buildStateMenu();
	
	// Build state panel
	JPanel statePanel = new JPanel();
	border(statePanel, "Current State");
	statePanel.add(stateMenu);
       	controlButton(RESET, statePanel);
	//statePanel.setMaximumSize(new Dimension(250, 60));

	// Build control panel
	JPanel controlPanel = new JPanel();
	controlPanel.add(Box.createRigidArea(new Dimension(0, 10)));
	controlButton(CLEAR, controlPanel);
	controlPanel.add(Box.createRigidArea(new Dimension(0, 10)));
	controlButton(SAVE, controlPanel);

	JPanel logPanel = new JPanel();
	logPanel.setLayout(new BoxLayout(logPanel, BoxLayout.Y_AXIS));
	border(logPanel, "Log of Events ( States )");
	JPanel logPanel2 = new JPanel();
	logPanel2.add(logScroll);
	logPanel.add(controlPanel);
	logPanel.add(logPanel2);

	// Add everything to main panel
	JPanel leftPanel = new JPanel();
        leftPanel.setLayout(new BoxLayout(leftPanel, BoxLayout.Y_AXIS));
	leftPanel.add(statePanel);
	leftPanel.add(Box.createRigidArea(new Dimension(0, 10)));
	leftPanel.add(signPanel);
       	leftPanel.add(Box.createRigidArea(new Dimension(0, 10)));	
	leftPanel.add(eventPanel);
	//sizePanel(topPanel, 950, 100);

	add(leftPanel);
	add(Box.createRigidArea(new Dimension(20, 0)));
	add(logPanel);
    }

    public void rebuildTab() {
	removeAll();
	build();
    }
    
    private void sizePanel(JPanel panel, int w, int h) {

	Dimension dim = new Dimension(w, h);
	panel.setMinimumSize(dim);
	panel.setPreferredSize(dim);
	panel.setMaximumSize(dim);
    }

    private void border(JComponent panel, String title) {
        Border e = BorderFactory.createEtchedBorder();
        Border t = BorderFactory.createTitledBorder(e, title);
        panel.setBorder(t);  	
    }
 

    public void controlButton(String label, JPanel panel) {
	JButton b = new JButton(label);
	b.addActionListener(this);
	b.setActionCommand(label);
	panel.add(b);
    }
    
    public void buildStateMenu() {
	stateMenu = new JComboBox(model().states());
	stateMenu.setBackground(Color.WHITE);
	stateMenu.addActionListener(new StateMenuListener());
	reset();
    }

    public void buildEventPanel() {
	
	String[] events = model().events();

        eventPanel = new JPanel();
	eventPanel.setLayout(new GridLayout(0, 6, 8, 8));
	border(eventPanel, "Events");

	eventButtons = new EventButton[events.length];
	ButtonListener eventListener = new ButtonListener();
	
	Dimension eDim = new Dimension(100,40);

	for (int i = 0; i < events.length; i++) {
            eventButtons[i] = new EventButton(events[i], i);
	    eventButtons[i].setMinimumSize(eDim);
	    eventButtons[i].setPreferredSize(eDim);
	    eventButtons[i].setMaximumSize(eDim);
            eventButtons[i].addActionListener(eventListener);
            eventPanel.add(eventButtons[i]);
        }

	
    }
    
    private void rebuildStateMenu() {
	listening = false;
	stateMenu.removeAllItems();
	String[] states = model().states();
	if (states != null) {
	    for (int i = 0; i < states.length; i++) {
		stateMenu.addItem(states[i]);
	    }
	}
	listening = true;
	reset();
    }

    private void rebuildEventPanel() {
	buildEventPanel();
	remove(1);
	add(eventPanel, 1);
    }

    private void reset() {
	int s = model().initialStateIndex();
	setState(s);
	log("[Set to " + model().states()[s] + "]\n"); 
    }

    // Reset button pressed
    public void actionPerformed(ActionEvent e) {

	String c = e.getActionCommand();

	if (c.equals(RESET)) {
	    reset();
	} else if (c.equals(CLEAR)) {
	    clearLog();
	} else if (c.equals(SAVE)) {
	    maui.operator().saveLog(log.getText());
	}
    }

    public void setState(int s) {
	
	listening = false;
      	stateMenu.setSelectedIndex(s);
	listening = true;

	enableButtons(s);
	showSigns(s);	
    }

    private void enableButtons(int s) {

	next = model().rangeIndices(s);

	for (int i = 0; i < next.length; i++) {
	    eventButtons[i].setEnabled(next[i] != null);
	}
    }

    private void showSigns(int s) {

	int[] indices = model().stateSignIndices(s);
	String[] names = new String[0];

	if (indices != null) {
	    names = new String[indices.length];
	    
	    for (int i = 0; i < indices.length; i++) {
		try {
		    names[i] = model().sign(indices[i]);
		} catch (FUIException e) {}
	    }
	  
	} 

	signs.setListData(names);
    }




    /*
     *  Logging
     */
    public void log(String pStr) {
        log.append(pStr);	
    }
  
    public void clearLog() {
        log.setText("");
	log("[Set to " +  model().states()[stateMenu.getSelectedIndex()] + "]\n");
    }

    /**
     * Button Presses
     */
    class ButtonListener implements ActionListener {

        public void actionPerformed(ActionEvent e) {

	    EventButton button = (EventButton) e.getSource();
	    int event = button.getIndex();
	    String eLabel = model().events()[event];

	    int[] to = next[event];
	    int s;

	    if (to != null) {
		if (to.length == 1) {
		    s = to[0];
		    setState(s);
		    log(eLabel + " (" + model().states()[s] + ")\n");
		} else {
		    s = to[random.nextInt(to.length - 1)];
		    setState(s);
		    log(eLabel + " (" + model().states()[s] + ")*\n");
		}
	    }
	}
    }

    class EventButton extends JButton {
	
	int index;

	public EventButton(String label, int i) {
	    super(label);
	    index = i;
	}

	public int getIndex() {
	    return index;
	}
    }

    class StateMenuListener implements ActionListener {

	public void actionPerformed(ActionEvent e) {
	    if (listening) {
		int s = stateMenu.getSelectedIndex();
		setState(s);
		log("[Set to " + model().states()[s] + "]\n"); 
	    }
	}
    }

}
